/*
 * EventTransport.h
 *
 * External (client-visible) interface to Transport Integration Layer (TIL)
 * plugin libraries within the Apama Integration Adapter Framework (IAF).
 * This is the interface that must be implemented by all transport libraries
 * in order to be successfully loaded and integrated by the IAF.
 *
 * $Copyright(c) 2002-2006, 2008, 2010 Progress Software Corporation (PSC). All rights reserved.$
 * $Copyright (c) 2013-2014, 2016-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 *
 * $Revision: 340684 $
 */

#ifndef AP_EVENT_TRANSPORT_H
#define AP_EVENT_TRANSPORT_H

/** \file 
 * 
 * External (client-visible) interface to Transport Integration Layer (TIL)
 * plugin libraries within the Apama Integration Adapter Framework (IAF).
 * This is the interface that must be implemented by all transport libraries
 * in order to be successfully loaded and integrated by the IAF.
 */


/*
 * Define this here so EventCodec.h picks it up as well
 */
typedef void* AP_TransportEvent;


#include <AP_Types.h>
#include <IAF_Platform.h>
#include <EventCodec.h>
#include <AP_TimestampSet.h>
#include <IAF_TimestampConfig.h>

/**
 * Transport API version.  The upper 16 bits identify the major release of the
 * API; the lower 16 bits indicate minor versions within that release.  No
 * changes that break backward binary compatibility will be made within a
 * given major release of the API.  That is, minor releases may add
 * functionality and fix bugs, but will not change or remove any documented
 * behaviour.
 *
 * Version history:
 * 1.0	  First public version, released with Apama 2.2.0
 * 1.1	  Add versioning and multiple transport support for Apama 2.2.1
 * 2.0	  Force users to rebuild for VS7.1
 * 2.1	  Fixes for 3.0 to correct bug 5063
 * 4.0	  Addition of high-accuracy timestamps for latency measurement.
 */
#define AP_EVENTTRANSPORT_VERSION	  0x00040000


/*
 * Names of functions exported from shared library.  The IAF will search for
 * these functions by name when the library is loaded, so you must use these
 * *exact* names when implementing a transport library.
 */
#define AP_EVENTTRANSPORT_INFO_FUNCTION_NAME AP_EventTransport_info
#define AP_EVENTTRANSPORT_CTOR_FUNCTION_NAME AP_EventTransport_ctor
#define AP_EVENTTRANSPORT_DTOR_FUNCTION_NAME AP_EventTransport_dtor


#ifdef __cplusplus
extern "C" {
#endif  /* __cplusplus */


/*
 * Forward-declare some structures
 */
struct AP_EventTransport;
struct AP_EventDecoder;


/**
 * AP_EventTransportError
 *
 * Error codes that can be returned by transport library functions.  The
 * enumeration values follow the normal UNIX convention of zero == OK and
 * non-zero == error.
 */
typedef enum {
	/// Everything is fine 
	AP_EventTransport_OK = 0,		   	
	/// Some unspecified internal error occurred 
	AP_EventTransport_InternalError,	
	/// Trouble reading/writing the external transport 
	AP_EventTransport_TransportFailure, 
	/// Trouble sending transport event to decoder 
	AP_EventTransport_DecodingFailure,  
	/// Transport was passed an invalid property set 
	AP_EventTransport_BadProperties,	
	/// Transport could not start correctly 
	AP_EventTransport_CantStart 		
} AP_EventTransportError;


/**
 * AP_EventTransportProperty
 *
 * A single transport property, corresponding to a &lt;property&gt; element in the
 * adapter configuration file.
 */
typedef struct {
	/// Transport property name
	AP_char8* name;
	/// Transport property value
	AP_char8* value;
} AP_EventTransportProperty;


/**
 * AP_EventTransportProperties
 *
 * Properties for the transport, extracted from the &lt;transport&gt; element in the
 * adapter configuration file.
 */
typedef struct {
	/// Transport property value
	AP_char8* name;
	/// Array of transport properties
	AP_EventTransportProperty** properties;
} AP_EventTransportProperties;


/**
 * AP_EventTransportStatus
 *
 * Transport status information structure, filled in by the getStatus call.
 * 
 * The first field (status) is a free-form text string that the transport 
 * can use to report any custom status information it might have. The 
 * iaf_watch tool will display the contents of this string. Note that the 
 * length of the status string is limited, currently to 1024 characters. 
 * Longer strings will be silently truncated. The next two fields 
 * (totalReceived and totalSent) report the total number of events received 
 * and sent by the transport. The last field, a pointer to an AP_NormalisedEvent, 
 * can contain custom information such as the state of the adapter.
 * 
 */
typedef struct {
	/**
	 * Free form text string containing a transport specific status message. 
	 * Strings longer than 1024 will be truncated
	 */
	AP_char8* status;
	/**
	 * Total number of downstream events received since IAF was run.
	 */
	AP_uint64 totalReceived;
	/**
	 * Total number of upstream events since the IAF was run
	 */
	AP_uint64 totalSent;
	/**
	 * Additional status information about this transport.
	 * 
	 * Standard  statusDictionary keys are 
	 * <pre><code>
	 * VERSION=transport_version_string
	 * CONFIG_VERSION=config_version_string
	 * CONNECTION=connectionGeneration (If the adapter manages only one connection)
	 * or
	 * CONNECTION_connectionId=connectionGeneration (If the adpater manages multiple connections)
	 * </code></pre>
	 */
	AP_NormalisedEvent* statusDictionary;
} AP_EventTransportStatus;



/**
 * AP_EventTransport_Functions
 *
 * Table of client visible functions exported by a transport library
 * instance.  These functions declare the only operations that may be
 * performed by users of a transport.
 *
 * Note that all of these functions take an initial AP_EventTransport*
 * argument; this is analogous to the (hidden) 'this' pointer passed to a C++
 * object when a member function is invoked on it.
 */
struct AP_EventTransport_Functions {

	/**
	 * updateProperties
	 *
	 * Update the configuration of the transport.  The transport may assume
	 * that stop(), flushUpstream() and flushDownstream() have all been called
	 * before this function is invoked.  The recommended procedure for
	 * updating properties is to first compare the new property set with the
	 * existing stored properties -- if there are no changes, no action should
	 * be taken.  Any pointer to the old property set becomes invalid as soon
	 * as this function returns; any such pointers should therefore be
	 * discarded in favour of the supplied new properties.
	 *
	 * @param transport The event transport instance
	 * @param properties The new transport property set derived from the IAF
	 * configuration file
	 * @param timestampConfig Timestamp recording/logging settings
	 * @return Event transport error code.  If this is not
	 * AP_EventTransport_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_EventTransportError (*updateProperties)(struct AP_EventTransport* transport, AP_EventTransportProperties* properties, IAF_TimestampConfig* timestampConfig);

	/**
	 * sendTransportEvent
	 *
	 * Called by an event encoder to send a message to the external transport.
	 * Ownership of the message is transferred to the transport when this
	 * function is called.  It is assumed that the encoder and transport share
	 * the same definition of the content of the event, so that the transport
	 * can effectively interpret the event and free any dynamically-allocated
	 * memory.
	 *
	 * @param transport The event transport instance
	 * @param event The event to be sent on the external transport. Ownership
	 * is transferred to the callee.
	 * @param timeStamp Timestamps associated with this event. Ownership is
	 * transferred to the callee.
	 * @return Event transport error code.  If this is not
	 * AP_EventTransport_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_EventTransportError (*sendTransportEvent)(struct AP_EventTransport* transport, AP_TransportEvent event, AP_TimestampSet* timeStamp);

	/**
	 * addEventDecoder
	 *
	 * Add a named event decoder to the set of decoders known to the
	 * transport.  If the named decoder already exists, it should be
	 * replaced.
	 *
	 * @param transport The event transport instance
	 * @param name The name of the decoder to be added
	 * @param decoder The decoder object itself
	 */
	void (*addEventDecoder)(struct AP_EventTransport* transport, const AP_char8* name, struct AP_EventDecoder* decoder);

	/**
	 * removeEventDecoder
	 *
	 * Remove a named event decoder from the set of decoders known to the
	 * transport.  If the named decoder does not exist, the function should do
	 * nothing.
	 *
	 * @param transport The event transport instance
	 * @param name The decoder to be removed
	 */
	void (*removeEventDecoder)(struct AP_EventTransport* transport, const AP_char8* name);

	/**
	 * flushUpstream
	 *
	 * Flush any pending normalized events onto the external transport. The
	 * transport may assume that the stop() function has been called before
	 * this function, so in many cases no action will be required to complete
	 * the flushing operation.
	 *
	 * @param transport The event transport instance
	 * @return Event transport error code.  If this is not
	 * AP_EventTransport_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_EventTransportError (*flushUpstream)(struct AP_EventTransport* transport);

	/**
	 * flushDownstream
	 *
	 * Flush any pending transport events into the decoder.  The transport may
	 * assume that the stop() function has been called before this function,
	 * so in many cases no action will be required to complete the flushing
	 * operation. Under no circumstances should any events be sent to the 
	 * Correlator after flushDownstream() has returned. 
	 *
	 * @param transport The event transport instance
	 * @return Event transport error code.  If this is not
	 * AP_EventTransport_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_EventTransportError (*flushDownstream)(struct AP_EventTransport* transport);

	/**
	 * start
	 *
	 * Establish a connection and start processing incoming data from the 
	 * external transport.
	 *
	 * When the start function is invoked the event transport is effectively 
	 * signaled to start accepting incoming messages and pass them onto a codec. 
	 * Events should not be sent to the correlator until the start function is called
	 * It is up to the event transport to determine which codec to communicate 
	 * with from the list of codecs made available to it through addEventDecoder 
	 * and removeEventDecoder. 
	 *
	 * Typically a configuration property would be used to specify the codec to 
	 * be used. If a handle to the desired codec had been stored in a variable 
	 * called decoder (of type AP_EventDecoder*) when addEventDecoder was called, 
	 * an event could be passed on to the codec using:
	 *
	 * decoder-&gt;functions-&gt;sendTransportEvent(decoder, event);
	 *
	 * An adapter should send events to the correlator only after its start() 
	 * method is called and before the stop() method returns. Therefore we 
	 * strongly recommend that a transport should not change to a state where 
	 * it is possible to receive events from any external transport until the 
	 * start() method has been called. In many cases, adapters will also need 
	 * to communicate with service monitors in the correlator to ensure that 
	 * the required monitors and event definitions are injected before they 
	 * begin to process messages from the external system. This is necessary in 
	 * order to avoid events from the adapter being lost if the correlator is 
	 * not yet ready to parse and process them.
	 *
	 * @param transport The event transport instance
	 * @return Event transport error code.  If this is not
	 * AP_EventTransport_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_EventTransportError (*start)(struct AP_EventTransport* transport);

	/**
	 * stop
	 *
	 * Stop processing incoming data from the external transport, typically 
	 * by pausing or closing down connections. 
	 *
	 * Adapter authors must ensure that no events are sent to the Correlator 
	 * after stop() has returned (the only exception being rare cases where the 
	 * transport sends buffered events in the Correlator in the 
	 * flushDownstream() method, which is called by the IAF after stop()). 
	 * If necessary any messages that are unavoidably received from the 
	 * transport after stop() has returned should be blocked, queued or simply 
	 * dropped.
	 *
	 * @param transport The event transport instance
	 * @return Event transport error code.  If this is not
	 * AP_EventTransport_OK, the getLastError() function should be called to
	 * get a more detailed description of what went wrong.
	 */
	AP_EventTransportError (*stop)(struct AP_EventTransport* transport);

	/**
	 * getLastError
	 *
	 * Return the transport's stored error message, if any.  The message
	 * string is owned by the transport so should not be modified or freed by
	 * the caller.
	 *
	 * @param transport The event transport instance
	 * @return The last error message generated by the transport
	 */
	const AP_char8* (*getLastError)(struct AP_EventTransport* transport);

	/**
	 * getStatus
	 *
	 * Fill in the supplied AP_EventTransportStatus structure with up-to-date
	 * status information for the transport.  Note that any data pointed to by
	 * the returned structure (such as strings) remains owned by the
	 * transport.  The caller must copy this data if it wishes to modify it.
	 *
	 * The AP_EventTransportStatus structure contains four fields. The first 
	 * field is a free-form text string that the transport can use to report 
	 * any custom status information it might have. The iaf_watch tool will 
	 * display the contents of this string. Note that the length of the 
	 * status string is limited, currently to 1024 characters. Longer strings 
	 * will be silently truncated. The next two fields report the total number 
	 * of events received and sent by the transport. The last field, a pointer 
	 * to an AP_NormalisedEvent, can contain custom information such as the 
	 * state of the adapter.
	 * 
	 * @param transport The event transport instance
	 * @param status The status structure to be filled in
	 */
	void (*getStatus)(struct AP_EventTransport* transport, AP_EventTransportStatus* status);
};


/**
 * AP_EventTransport
 *
 * External (client-visible) interface to an IAF TIL plugin library.  The
 * AP_EventTransport struct contains a table of function pointers, declared in
 * the AP_EventTransport_Functions struct above.  The implementation of these
 * functions is private.  Users of the transport libary should invoke
 * functions on it as in the following example (transport is of type
 * AP_EventTransport*):
 *
 * i = transport-&gt;functions-&gt;sendTransportEvent(mapper, event, timestamps);
 */
struct AP_EventTransport {

	/**
	 * Pointer to private internal data.
	 */
	void* reserved;

	/**
	 * Function table of transport library operations.  See documentation for
	 * AP_EventTransport_Functions for details.
	 */
	struct AP_EventTransport_Functions* functions;
};
typedef struct AP_EventTransport AP_EventTransport;


/**
 * AP_EventTransportInfoPtr
 *
 * Pointer to the info function for the tranport library.  Each transport
 * plugin library must export a function with this signature, named using the
 * AP_EVENTTRANSPORT_INFO_FUNCTION_NAME macro.
 *
 * Returns basic information about the transports implemented by this shared
 * library, along with the transport API version it was built against.
 *
 * @param version 'Out' parameter for version number of transport library.
 * This should be the value of AP_EVENTTRANSPORT_VERSION that was defined in
 * EventTransport.h when the transport was built.  The IAF will check the
 * version number to ensure that it can support all the capabilities offered
 * by the transport
 * @param capabilities 'Out' parameter for the capabilities of the transports
 * provided by this library.  Currently no such capabilities are defined, so
 * the value of this field will be ignored by the IAF.  However, all bits of
 * the capabilities value are reserved for future use, so current transports
 * should ensure that they set this value to zero.
 */
typedef AP_EVENTTRANSPORT_API void (AP_EVENTTRANSPORT_CALL* AP_EventTransportInfoPtr)(AP_uint32* version, AP_uint32* capabilities);


/**
 * AP_EventTransportCtorPtr
 *
 * Pointer to the constructor function for the transport library.  Each TIL
 * plugin library must export a function with this signature, named using the
 * AP_EVENTTRANSPORT_CTOR_FUNCTION_NAME macro.
 *
 * Constructs a new instance of the event transport.  This function will be
 * called by the adapter main program when the adapter starts up.  The
 * transport should be created in a 'stopped' state, such that it is not
 * actively processing data from the external transport.  A call to the
 * start() function is required to begin processing.
 *
 * Note that the input parameters (name, properties, timestampConfig) are owned
 * by the caller.  The IAF framework guarantees that propertes and
 * timestampConfig will remain valid until after a subsequent call to
 * updateProperties(), so it is safe to hold a pointer to this structure.
 * You should, however, copy the name string if you wish to keep it.  The
 * contents of the output parameter errMsg belongs to the transport.
 *
 * @param name The name of this transport instance
 * @param properties Transport property set derived from the IAF config file
 * @param err 'Out' parameter for error code if constructor fails
 * @param errMsg 'Out' parameter for error message if constructor fails
 * @param timestampConfig Timestamp recording/logging settings
 * @return Pointer to a new transport instance or NULL if the constructor
 * fails for some reason.  In the case of failure, err and errMsg should be
 * filled in with an appropriate error code and message describing the
 * failure.
 */
typedef AP_EVENTTRANSPORT_API AP_EventTransport* (AP_EVENTTRANSPORT_CALL* AP_EventTransportCtorPtr)(AP_char8* name, AP_EventTransportProperties* properties, AP_EventTransportError* err, AP_char8** errMsg, IAF_TimestampConfig* timestampConfig);

/**
 * AP_EventTransportDtorPtr
 *
 * Pointer to the destructor function for the transport library.  Each TIL
 * plugin library must export a function with this signature, named using the
 * AP_EVENTTRANSPORT_DTOR_FUNCTION_NAME macro.
 *
 * Destroys an instance of the event transport that was previously created by
 * AP_EventTransport_ctor.  The transport may assume that it has been stopped
 * and flushed before the destructor is called.
 *
 * @param transport The event transport object to be destroyed
 */
typedef AP_EVENTTRANSPORT_API void (AP_EVENTTRANSPORT_CALL* AP_EventTransportDtorPtr)(AP_EventTransport* obj);


#ifdef __cplusplus
}  /* extern "C" */
#endif  /* __cplusplus */


#endif  /* AP_EVENT_TRANSPORT_H */
