/*
 * Title:        bits/buffer_t.hpp
 * Description:  C++ header-only wrapper for C-ABI data_t type
 * $Copyright (c) 2015-2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: buffer_t.hpp 320837 2017-11-30 21:07:30Z bsp $
 */

/* This file:
 * Contains the C++ implementation of buffer_t
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

/**
 * A class that holds an untyped byte buffer. Iterators and indexing give access
 * to the underlying byte[] which is guaranteed to be contiguous and at least as
 * large as size().
 *
 * This is moveable, but not implicitly copyable. If you want to copy it use
 * the .copy() method. If you want to move it use std::move().
 */
class buffer_t: protected sag_underlying_buffer_t
{
	friend class data_t;
public:
	/** Construct an empty buffer */
	buffer_t()
	{
		table = nullptr;
	}
	/** Construct a buffer with (at least) capacity for n items */
	explicit buffer_t(size_t n);

	/** Free the underlying buffer contents */
	~buffer_t()
	{
		clear();
	}

	/** Move assignment. Other will be left as the empty buffer and our previous
		contents will be freed. */
	buffer_t &operator=(buffer_t &&other)
	{
		buffer_t tmp;
		// put us in tmp
		swap(std::move(tmp));
		// put other in us (and by extension the null tmp in other)
		swap(std::move(other));
		return *this;
	}
	/** Move constructor. Other will be left as the empty buffer */
	buffer_t(buffer_t &&other)
	{
		table = nullptr;
		swap(std::move(other));
	}
	/** Swap the contents of this buffer and other. */
	void swap(buffer_t &&other)
	{
		std::swap(table, other.table);
	}

	/** Return a deep copy of this buffer */
	buffer_t copy() const;

private:
	/** This is not implicitly copiable. Use the copy() method. */
	buffer_t &operator=(const buffer_t &other) { abort(); }
	/** This is not implicitly copiable. Use the copy() method. */
	buffer_t(const buffer_t &other) { abort(); }
public:

	typedef uint8_t *iterator;
	typedef const uint8_t *const_iterator;

	/** Returns true if both buffers are the same size and have the same contents. */
	bool operator==(const buffer_t &other) const
	{
		return (this->size() == other.size() &&
			std::equal(this->cbegin(), this->cend(), other.begin())
		);
	}
	/** Returns true if both buffers are different sizes or have different contents. */
	bool operator!=(const buffer_t &other) const
	{
		return !(this->operator==(other));
	}

	/* Iteration: 
	 * begin() and end() methods, both const and non-const, forward and backward.
	 */
	/** Forward iterator begin */
	iterator begin() { return table?table->data:0; }
	/** Forward iterator end */
	iterator end() { return table?table->data+table->length:0; }
	/** Forward const_iterator begin */
	const_iterator begin() const { return table?table->data:0; }
	/** Forward const_iterator end */
	const_iterator end() const { return table?table->data+table->length:0; }
	/** Forward const_iterator begin */
	const_iterator cbegin() const { return table?table->data:0; }
	/** Forward const_iterator end */
	const_iterator cend() const { return table?table->data+table->length:0; }

	/** Returns the size of the buffer */
	size_t size() const { return table ? table->length : 0; }
	/** Returns true if the buffer is unallocated (size() == 0) */
	bool empty() const { return 0 == size(); }

	/** Return a reference to the byte at the given offset.
		If the offset is outside the buffer then throws std::runtime_error */
	uint8_t &operator[](size_t n)
	{
		if (n >= size()) throw std::runtime_error("Out of bounds");
		return table->data[n];
	}
	/** Return a reference the byte at the given offset.
		If the offset is outside the buffer then throws std::runtime_error */
	const uint8_t &operator[](size_t n) const
	{
		if (n >= size()) throw std::runtime_error("Out of bounds");
		return table->data[n];
	}
	/**
	 * Free the underlying buffer;
	 */
	void clear();
};

