/*
 * Title:			bits/map_extractor.hpp
 * Description:		Helper class for extracting values from the map_t type
 * $Copyright (c) 2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 */

/* This file:
 * Contains the C++ implementation of MapExtractor
 *
 */

#ifndef _DATAT_BITS_INCLUDE_TAG
#error Must only be included from sag_connectivity_cpp.hpp, use that instead
#endif

/**
 * Provides a compile-time type-safe way to extract values from a map_t, with user-friendly error messages identifying the map and key within the map if any item is missing.
 *
 * Useful for handling message payload, metadata or plug-in configuration objects.
 * Typical usage would involve creating an instance for accessing a plug-in's configuration, calling the appropriate get functions to extract all required information, before calling the checkNoItemsRemaining() function to produce a clear error if the user specified any keys that were not expected.
 */
class MapExtractor
{
public:
	/** 
	 * Construct a MapExtractor wrapping the specified map. 
	 * The lifetime of this object must be shorter than the lifetime of the wrapped map.
	 */
	MapExtractor(const map_t &map, const std::string &displayName)
		: map(map), keysAlreadyAccessed(), displayName(displayName)
	{
		if (displayName.empty())
		{
			throw std::runtime_error("displayName cannot be empty");
		}
	}

	MapExtractor(const MapExtractor &) = delete;
	MapExtractor &operator=(const MapExtractor &) = delete;
	MapExtractor &operator=(MapExtractor &&) = delete;
	~MapExtractor() = default;

	/** Construct a MapExtractor, moving the contents from another MapExtractor. */ 
	MapExtractor(MapExtractor &&other) : 
		map(other.map), 
		keysAlreadyAccessed(std::move(other.keysAlreadyAccessed)), 
		displayName(std::move(other.displayName))
	{}

	/** Returns true if the map is empty. */ 
	bool empty()
	{
		return map.empty();
	}

	/** Returns a reference to the wrapped map. */ 
	const map_t& getUnderlyingMap()
	{
		return map;
	}

	/** Returns the display name associated with this map. */
	std::string getDisplayName()
	{
		return displayName;
	}

	/** 
	 * Returns a string representation of this object, including both the display name and map contents.
	 * Intended for diagnostic purposes and not for identitifaction; the format of this string may change in the future.
	 */
	std::string toString();

	/**
	 * Checks that all entries in this map have been read (using one of the get functions in this class), throwing an exception listing any unexpected items.
	 */ 
	void checkNoItemsRemaining()
	{
		std::vector<std::string> remaining;
		for (auto it = map.begin(); it != map.end(); ++it)
		{
			if (keysAlreadyAccessed.find(it->first) == keysAlreadyAccessed.end())
			{
				remaining.push_back(stringifyKey(it->first));
			}
		}
		if (remaining.empty()) return;

		std::stringstream ss;
		for (auto it = remaining.begin(); it != remaining.end(); ++it)
		{
			ss << "'" << *it << "'" << " ";
		}
		throw std::runtime_error("Found unexpected items in the map '" + displayName + "': " + ss.str());
	}

	/**
	 * \internal
	 * Returns a value of type T from the map, throwing an exception if the key is missing or the associated value is the empty data_t. 
	 *
	 * This specialization works for types which are not convertable to/from string and just calls the underlying get method, returning
	 * a reference.
	 *
	 * The publically documented version works with convertable to/from string versions and will convert if neccessary. Note that to do
	 * conversions to string you have to use std::string and not const char *.
	 */ 
	template<typename T>
	const is_not_convertable_t<T> &get(const data_t& key)
	{
		return com::softwareag::connectivity::get<T>(doGet(key));
	}

	/**
	 * Returns a value of type T from the map, throwing an exception if the key is missing or the associated value is the empty data_t. 
	 * 
	 * If the type T is one of double, int64_t, bool or std::string then the underlying value will automatically be converted to/from
	 * the string form as appropriate to convert it to the requested type. In this case the return is by value and is a copy of the 
	 * underlying value. An exception will be thrown if the actually underlying data is not a type which can be converted to the requested
	 * type.
	 *
	 * If the type T is one of the other types which can be stored in a data_t (including const char *) then no conversion will be
	 * attempted. In this case the return is by const reference to the underlying value. An exception will be thrown if the underlying
	 * data is not the requested type.
	 */ 
	template<typename T>
	const is_convertable_t<T> get(const data_t& key)
	{
		return com::softwareag::connectivity::convert_to<T>(doGet(key));
	}

	/**
	 * Returns a value of type T from the map, returning the default value if the key is missing or the associated value is the empty data_t.
	 *
	 * T can only be one of the primitive types contained within a data_t (including const char *). The return is by value as a copy of 
	 * either the underlying value or of the defaultValue if the requested key is not in the map. The default value is also returned
	 * if the key is in the map, but the value is not of the requested type or convertable to the requested type.
	 */ 
	template<typename T>
	const T get(const data_t& key, const T defaultValue)
	{
		try
		{
			return get<T>(key);
		}
		catch (const std::exception &)
		{
			return defaultValue;
		}
	}


	/**
	 * Returns a MapExtractor wrapping a map_t from the map.
	 * If the key is missing or the associated value is the empty data_t and emptyIfMissing is true, returns a MapExtractor holding an empty map, else throws an exception.
	 */ 
	MapExtractor getMap(const data_t& key, bool emptyIfMissing)
	{
		try 
		{	
			return MapExtractor(get<map_t>(key), displayName + "." + stringifyKey(key));
		} 
		catch (const std::exception &) 
		{
			if (emptyIfMissing)
			{
				return MapExtractor(emptyMap, displayName + "." + stringifyKey(key));
			}
			throw;
		}
	}

	/**
	 * Returns a string value, throwing an exception if the key is missing, the associated value is the empty data_t or the string is empty.
	 */
	std::string getStringDisallowEmpty(const data_t& key)
	{
		std::string ret = get<const char*>(key);
		if (ret.empty())
		{
			throw std::runtime_error("The string value associated with key '" + stringifyKey(key) + "' in " + displayName + " is empty");
		}
		return ret;
	}

	/**
	 * Returns a string value, returning the defaultValue if the key is missing, the associated value is the empty data_t or the string is empty.
	 */
	std::string getStringDisallowEmpty(const data_t& key, std::string defaultValue)
	{
		try {
			return getStringDisallowEmpty(key);
		} catch (const std::exception &) {
			return defaultValue;
		}
	}

	/**
	 * Returns a string value, throwing an exception if the key is missing or the associated value is the empty data_t.
	 */
	std::string getStringAllowEmpty(const data_t& key)
	{
		return get<const char*>(key);
	}

	/**
	 * Returns a string value, returning the defaultValue if the key is missing or the associated value is the empty data_t.
	 */
	std::string getStringAllowEmpty(const data_t& key, std::string defaultValue)
	{
		try {
			return getStringAllowEmpty(key);
		} catch (const std::exception &) {
			return defaultValue;
		}
	}

private:
	/**
	 * Retrieve a value from the map, marking that this key has been accessed. Throws if the key is missing or the associated value is empty.
	 */ 
	const data_t& doGet(const data_t& key)
	{
		keysAlreadyAccessed[key.copy()] = data_t();

		auto it = map.find(key);
		if (it == map.end())
		{
			throw std::runtime_error("Key '" + stringifyKey(key) + "' not found in " + displayName);
		}
		if (it->second.type_tag() == SAG_DATA_EMPTY)
		{
			throw std::runtime_error("Key '" + stringifyKey(key) + "' has an empty value in " + displayName);
		}
		return it->second;
	}

	/** Helper function for stringifying a data_t. */
	std::string stringifyKey(const data_t& data);

	/** The underlying map this object wraps. */
	const map_t& map;
	/** The set of keys that have been accessed. */
	map_t keysAlreadyAccessed;
	/** The user-visible identifier for this object. */
	const std::string displayName;
	/** An empty map returned by the getMap() function. */
	const map_t emptyMap;
};
