/*
 * $Copyright (c) 2015-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: sag_connectivity_chain_managers.hpp 328256 2018-05-01 11:04:06Z antm $
 */


#ifndef _SAG_CONNECTIVITY_CHAIN_MANAGERS_HPP_
#define _SAG_CONNECTIVITY_CHAIN_MANAGERS_HPP_

#include <sag_connectivity_plugins.hpp>
#include <memory>
#include <unordered_map>
#include <vector>

/** \file 
 * API for transport connectivity plug-ins that can create chains. 
 */

namespace com {
namespace softwareag {
namespace connectivity { 
namespace chainmanagers {

/**
 * API for Chain Manager to listen for changes in channel subscriptions within the correlator.
 *
 * The chain manager will need to register listeners for interested channels by using the AbstractChainManager.host. 
 *
 * Thread-safety: The callbacks methods can be called from multiple threads so plug-in should take care of 
 * accessing any mutable state in thread safe manner.
 * 
 * The hostwards channels subscriptions in persistent monitor are persisted so listener will be notified of
 * channel creations of matching channels in HOSTWARDS direction during recovery. The transportwards channels
 * are not persisted so listener will not receive any channel creation notification in TRANSPORTWARDS direction
 * during recovery, instead it will be notified of a matching channel creation when first event is sent to the channel.
 * Direction is either TOWARDS_TRANSPORT or TOWARDS_HOST.
 *
 *	@since 10.2.0.0
 *
 */
class ChannelLifecycleListener
{
public:
	/** Virtual destructor. */
	virtual ~ChannelLifecycleListener() {}
	/**
	 * Called by host when a channel is created which starts with the prefix provided during listener registration.
	 *
	 * This method might be called multiple times for the same channel and direction, to give the manager a 
	 * chance to re-evaluate whether a chain is required for the given channel if the channel cache is flushed. 
	 * When this method is called the manager must first decide whether it currently needs to have a chain for the 
	 * specified channel, as some managers may only wish to take action if a channel with the specified name exists on 
	 * the external system they are connected to.
	 *
	 * The manager must also check if it already has a chain for this channel in the specified direction,
	 * since in some situations the listener will notify about creation of the same channel more than once.
	 *
	 * If the manager has established that a chain is needed for this channel and none already exists it should create
	 * one, if a chain already exists for this channel but is no longer needed it should destroy it, and otherwise
	 * it should do nothing.
	 *
	 * In the case where channel is created just after its destruction, it might be possible that onChannelCreated
	 * is called without onChannelDestroyed being called. So don't rely on the onChannelCreated and onChannelDestroyed invocations being paired.
	 *
	 * Any transportwards event sent to a channel by EPL while onChannelCreated call is in progress will be delivered
	 * to any chain that is started by the listener within that call and is subscribed to the channel. 
	 * Similarly any hostwards event sent sent be a chain started in onChannelCreated call to the channel created will be delivered 
	 * to the monitor that subscribed to the channel.
	 *
	 * Thread-safety: Any chain started in onChannelCreated should not send messages to host in start method of transport as it can cause deadlock because
	 * this method can be called from receiving thread. The transport can instead send the message from hostReady method which is always called from background thread.
	 * The method can be called from multiple threads so plug-in should take care of accessing any mutable state in a thread safe manner.
	 *
	 * The ChannelLifecycleListener's onChannelCreated and onChannelDestroyed methods will not be called once the 
	 * chain manager's shutdown() has begun, and any in-progress calls will be completed before shutdown() is invoked. 
	 *
	 * @param channel the name of the channel created
	 * @param direction the direction of channel, which is either towards the host or towards the transport. 
	 */
	virtual void onChannelCreated(const std::string &channel, Direction direction) = 0;

	/**
	 * Called by host when a channel is destroyed which starts with the prefix provided during listener registration.
	 *
	 * The method can be called from multiple threads so plug-in should take care of accessing any mutable state in
	 * thread safe manner.
	 *
	 * At present, channel destroy notifications are only sent for the TOWARDS_HOST direction
	 * (monitor.subscribe()) since in the TOWARDS_TRANSPORT direction (send ... to ...) there is no unambiguous way of
	 * determining when a channel is no longer needed. 
	 *
	 * The ChannelLifecycleListener's onChannelCreated and onChannelDestroyed methods will not be called once the 
	 * chain manager's shutdown() has begun, and any in-progress calls will be completed before shutdown() is invoked. 
	 *
	 * @param channel the name of channel destroyed
	 * @param direction the direction of channel, which is either towards the host or towards the transport. 
	 */
	virtual void onChannelDestroyed(const std::string &channel, Direction direction) = 0;
};

template<typename TRANSPORT>
class AbstractChainManager; // forward decl for Host friend

template<typename TRANSPORT>
class ChainManagerHost; // forward decl for Chain friend

/** A plug-in configuration.
 * This is a string name and a map_t for the configuration.
 * Note that unlike map_t, this is copyable - copies are deep copies. (needed to make ChainDefinition copyable)
 * @since 10.2.0.0
 */
struct PluginConfiguration {
	std::string pluginName;
	map_t config;
	PluginConfiguration(const std::string &pluginName, const map_t &config) : pluginName(pluginName), config(config.copy()) {}
	PluginConfiguration(const PluginConfiguration &other) : pluginName(other.pluginName), config(other.config.copy()) {}
	PluginConfiguration &operator=(const PluginConfiguration &other) {
		if(this != &other) {
			pluginName = other.pluginName;
			config=other.config.copy();
		}
		return *this;
	}
	PluginConfiguration(PluginConfiguration&& other) = default; /// while we override the copy behaviour, no change to move behaviour
	PluginConfiguration &operator=(PluginConfiguration &&other) = default; /// while we override the copy behaviour, no change to move behaviour
	~PluginConfiguration() = default; /// while we override the copy behaviour, no change to destructor

	/** @return The name of the plug-in. */
	const std::string& getPluginName() const { return pluginName; }
	
	/** @return The configuration for this plug-in. */
	const map_t &getConfiguration() const { return config; }
};


/**
* A dynamic chain definition, providing the configuration for each plug-in in a chain. Can be used by a manager to
* create a chain.
*
* This object is not thread-safe so synchronization should be used by the caller if it is to be mutated.
*
* It is possible for a chain manager to modify a chain definition before using it to create a chain,
* though it is usually recommended to use an existing dynamic chain definition from the configuration
* without modification instead (that is, except for any string substitutions passed into the createChain call).
* @since 10.2.0.0
*
*/
struct ChainDefinition : public std::vector<PluginConfiguration> {
	typedef std::vector<PluginConfiguration> vector_t;
	ChainDefinition() {}
	/**
	* @return the configuration associated with the transport at the end of this chain definition.
	* This is a reference not a copy so any changes made to the returned value will affect the chain definition.
	*/
	const map_t& getTransportConfig() const {
		return back().getConfiguration();
	}
};

typedef std::unordered_map<std::string, ChainDefinition> ChainDefinitions;

namespace converters_impl {
ChainDefinitions convertChainDefinitions(const map_t &chainDefinitions);
list_t convertChainDefinition(const ChainDefinition& chainDefinition);
}


/**
* Represents a dynamic chain instance, created using ChainManagerHost.createChain.
* @since 10.2.0.0
* 
* @author BSP
*/
template<typename TRANSPORT>
class Chain
{
	// so host can call the private constructor
	friend class ChainManagerHost<TRANSPORT>;
	// not copyable - this is a singleton ptr, but does not own the lifetime.
	Chain(const Chain &other) = delete;
	Chain &operator=(const Chain &other) = delete;
	Chain(const Chain &&other) = delete;
	Chain &operator=(const Chain &&other) = delete;
	void startImpl() const;
public:
	~Chain() = default; // default destructor is OK, though @private
	/**
	* Calls start on all of the plug-ins in the chain and connects the chain to receive events.
	*
	* Will not return until all start() methods have completed and therefore may block.
	*
	* If the chain to be created might send events during the start() call then this should not be
	* called from an event receiving thread as a deadlock may occur.
	*
	* If onApplicationInitialized has already been called then hostReady will be called on the new chain immediately
	* from a background thread. Once onApplicationInitialized has been called the chain will receive any events that
	* were sent to its subscribed channels from after the createChain call onwards.
	*
	* @return this Chain instance (to support fluent call style)
	* @throws Exception if an exception is thrown by the start() method of any plug-in in the chain.
	*/
	const Chain &start() const {
		startImpl();
		return *this;
	}
	/**
	* Calls start on all of the plug-ins in the chain and connects the chain to receive events.
	*
	* Will not return until all start() methods have completed and therefore may block.
	*
	* If the chain to be created might send events during the start() call then this should not be
	* called from an event receiving thread as a deadlock may occur.
	*
	* If onApplicationInitialized has already been called then hostReady will be called on the new chain immediately
	* from a background thread. Once onApplicationInitialized has been called the chain will receive any events that
	* were sent to its subscribed channels from after the createChain call onwards.
	*
	* @return this Chain instance (to support fluent call style)
	* @throws Exception if an exception is thrown by the start() method of any plug-in in the chain.
	*/
	Chain &start() {
		startImpl();
		return *this;
	}
	/** Shutdown and destroy a chain. This will result in the transport and any codecs being shutdown and then deleted.  */
	void destroy();

	/** Get the transport plug-in at the end of the chain. Never returns null. It is an error to call this after the chain has been destroyed. */
	TRANSPORT *getTransport() const { 
		if (!transport) throw std::runtime_error("Cannot call getTransport after chain has been destroyed");
		return transport; 
	}

	/** Get the unique identifier of this chain */
	const std::string &getChainId() { return chainId; }

	typedef std::shared_ptr<Chain> ptr_t;


private:
	Chain(TRANSPORT* transport, void* connectivityManager, const std::string &chainId, const std::string &managerName)
		:transport(transport), connectivityManager(connectivityManager), chainId(chainId), managerName(managerName)
	{}

	TRANSPORT* transport;
	void* connectivityManager;
	std::string chainId;
	// just for error messages for now
	std::string managerName;
};

/**
 * API provided to Chain Managers for callbacks into the host, and for operations such as creating chains and 
 * registering channel lifecycle listeners.
 * @since 10.2.0.0
 */
template<typename TRANSPORT>
class ChainManagerHost
{
	// give friend status to ChainManager so it can get the hostptr
	friend class AbstractChainManager<TRANSPORT>;

	/** @internal Create a Host from an underlying ptr */
	ChainManagerHost(void *connectivityManager, const AbstractChainManager<TRANSPORT> *chainManager): connectivityManager(connectivityManager), chainManager(chainManager) {}

public:
	/**
	 * Create a chain instance programmatically.
	 * @param chainInstanceIdSuffix A string identifier which will be suffixed onto "managerName-" to uniquely 
	 * identify the new chain instance.
	 * Caution: A small amount of memory is allocated for each unique chain instance identifier. This memory is not freed
	 * when the chain is destroyed. Therefore, if you are creating many chains, consider reusing old chain instance identifiers.
	 * If you create more than 1000 unique identifiers, a warning is written to the correlator log file to notify you of this. You
	 * cannot have two active chains with the same chain instance identifier, so only reuse identifiers which are no longer in use.
	 * @param defaultChannelTowardsHost Default channel to use for sending a message towards the host if no channel is 
	 * specified on the message; use empty string for no default. It is an error to specify a non-empty defaultChannelTowardsHost 
	 * value if defaultChannel is also specified for the host plug-in.
	 * @param subscribeChannels The set of channels that the chain should be subscribed to (for messages from the host to the transport). 
	 * @param chainDefinition The definition of the chain (passed to the manager at construction time)
	 * @param substitutions A string/string map of @{..} variables to replace in the chainDefinition
	 * @param transportConstructorArgs Zero or more arguments to be passed to the transport constructor as part of chain creation. The order and type 
	 * of arguments must match those in the transport constructor (after the ManagedTransportConstructorParameters argument). 
	 * @return a pointer to the newly created transport instance??? or is it a reference to the chain containing that transport reference???. The lifetime of this transport 
	 * is managed by the connectivity plug-ins framework, so if the chain manager retains any references to the 
	 * transport it must ensure they are removed during transport shutdown or else at the very start of the transport's destructor. 
	 */
	template<typename ... ARGS>
	typename Chain<TRANSPORT>::ptr_t createChain(const std::string &chainInstanceIdSuffix, const std::string &defaultChannelTowardsHost, const list_t &subscribeChannels, const ChainDefinition &chainDefinition, const map_t &substitutions, ARGS ... transportConstructorArgs) const;

	/** Register a callback handler for subscribe/unsubscribe and send-to notifications.
	 *
	 * This method does not take ownership of the specified listener, but the listener must not be deleted 
	 * until after the chain manager has been shutdown.
	 * Typically the ChainManager class will itself implement the listener interface, 
	 * in which case the listener's lifetime will be correct without further action. 
	 *
	 * @param listener The listener which will get the notification when a channel starting with the channelPrefix is created or destroyed.
	 * @param channelPrefix The channel prefix for which listener should get callback.
	 *
	 */
	void addChannelLifecycleListener(ChannelLifecycleListener *listener, const std::string &channelPrefix) const;
private:

	template<typename TUPLE>
	typename Chain<TRANSPORT>::ptr_t createChainImpl(const std::string &chainInstanceIdSuffix, const std::string &defaultChannelTowardsHost, const list_t &subscribeChannels, const ChainDefinition &chainDefinition, const map_t &substitutions, const TUPLE& tuple) const;

	/// @internal a callback on this side of the DLL boundry which can be used to delete a listener
	static sag_error_t deleteChannelLifecycleListener(sag_plugin_t callback);
	/// @internal a callback on this side of the DLL boundry which can be used to call onChannelCreated
	static sag_error_t callOnChannelCreated(const char *managerLogName, sag_plugin_t callback, const char*, bool);
	/// @internal a callback on this side of the DLL boundry which can be used to call onChannelDestroyed
	static sag_error_t callOnChannelDestroyed(const char *managerLogName, sag_plugin_t callback, const char*, bool);
	/// @internal a bare pointer to the internal singleton which these are callbacks for
	void *connectivityManager;
	/// @internal a bare pointer to the manager which these are callbacks for
	const AbstractChainManager<TRANSPORT> *chainManager;
};


/**
* A container for parameters passed to the constructor of a managed transport (i.e. a transport that has a chain manager).
*
* Compatibility note: new methods may be added to this interface in future releases.
*
* The transport's constructor is called by the connectivity framework on creation of a chain, and receives 
* a ManagedTransportConstructorParameters and any additional transport args passed to the createChain method. 
* @since 10.2.0.0
*/
class ManagedTransportConstructorParameters : public PluginConstructorParameters::TransportConstructorParameters
{
public:
	/** Get the channels that the chain containing this transport instance is subscribed to for receiving 
	 *  messages from the host towards the transport. 
	 *  
	 *  @return the list of channels, or empty if this chain is only sending towards the host and not towards the transport.  
	 */
	const list_t &getSubscribeChannels() const { return subscribeChannels; }

	/**
	 * Get the default channel used on this chain for messages sent from the transport towards the host.
	 * @return the channel name, or an empty string if this chain is only sending towards the transport (or if no default was specified 
	 * when the chain was created by the manager).
	 */
	const std::string &getDefaultChannelTowardsHost() const { return defaultChannelTowardsHost; }

	/** Constructor - for internal use only
	@private
	*/
	ManagedTransportConstructorParameters(const std::string &pluginName, const std::string &chainId, const map_t &config, void* connectivityManager,
		const std::string &defaultChannelTowardsHost, const list_t subscribeChannels, void* reserved)
		: TransportConstructorParameters(pluginName, chainId, config, connectivityManager, reserved),
		defaultChannelTowardsHost(defaultChannelTowardsHost), subscribeChannels(subscribeChannels.copy())
	{}

private:
	std::string defaultChannelTowardsHost;
	list_t subscribeChannels;
};

/**
* A container for parameters passed to the constructor of a dynamic chain manager.
*
* Compatibility note: new methods may be added to this interface in future releases.
* @since 10.2.0.0
*/
class ChainManagerConstructorParameters
{
	template <typename TRANSPORT> friend class AbstractChainManager;
	// not copyable: ChainManagers should use the config, etc that are their field members.
	ChainManagerConstructorParameters(const ChainManagerConstructorParameters &params) = delete;
	ChainManagerConstructorParameters& operator=(const ChainManagerConstructorParameters &params) = delete;
public:
	/** Get the configuration for this plug-in. 
	 *
	 * It is usually more convenient to access configuration using the non-const (mutable) config field of AbstractChainManager instead of this method. 
	 */
	const map_t &getManagerConfig() const { return managerConfig; }

	/**
	 * @return the name used in the configuration file for this dynamic chain manager.
	 */
	const std::string &getManagerName() const { return managerName; }

	/**
	 * @return the name used in the configuration file for the transport plug-in that is managed by this chain manager.
	 */
	const std::string &getTransportPluginName() const { return transportPluginName; }

	/** Constructor - for internal use only
	@private
	*/
	ChainManagerConstructorParameters(const map_t &managerConfig, const std::string &managerName, const std::string &transportPluginName, void* connectivityManager, const map_t &chainDefs, void* reserved)
		: managerName(managerName), transportPluginName(transportPluginName), managerConfig(managerConfig), chainDefs(chainDefs), connectivityManager(connectivityManager), reserved(reserved)
	{}

private:
	std::string managerName;
	std::string transportPluginName;
	const map_t &managerConfig;
	const map_t &chainDefs;
	/** For status reporting */
	void* connectivityManager;
	/** Reserved for future use, just in case we unavoidably have to add something without breaking binary compat */
	void* reserved;
};



/**
 * Base class for transport chain manager plug-ins.
 *
 * The manager will only be shutdown and destroyed on correlator shutdown. On correlator shutdown, the order of callbacks is:
 * <ul>
 * <li> ChainManager.shutdown() called.  This is the last point a chain manager can safely call destroy() on any Chains it has created.
 * <li> Transport.shutdown() called on any remaining transports.
 * <li> Transports destroyed (destructor called) on any remaining transports.
 * <li> ChainManager destroyed (destructor called).
 * </ul>
 * The manager can thus call Chain.destroy() during shutdown, but not after, and is not required to.
 * To export a transport chain manager class for use in the host you should use the SAG_DECLARE_CONNECTIVITY_TRANSPORT_CLASS_CHAIN_MANAGER(Class) macro ; the transport should not be exported
 * The ChainManager must provide a public constructor with the same signature as
 * AbstractChainManager(const ChainManagerConstructorParameters &params)
 *	@since 10.2.0.0
 */
template<typename TRANSPORT>
class AbstractChainManager
{

public:

	/** The transport that this host can create chains for */
	typedef TRANSPORT transport_t;

	/** Give the manager instance name, a host object for callbacks, the configuration specified in managerConfig and the map of chain definitions passed to the manager constructor. */
	AbstractChainManager(const ChainManagerConstructorParameters &params)
		: 
		logger("connectivity." + params.getManagerName() + "." + (params.getManagerName() == params.getTransportPluginName() ? "manager" : params.getTransportPluginName())),
		managerName(params.getManagerName()), transportPluginName(params.getTransportPluginName()),
		chainDefinitions(std::move(converters_impl::convertChainDefinitions(params.chainDefs))), 
		host(params.connectivityManager, this),
		config(params.getManagerConfig().copy()),
		statusReporter(new StatusReporter(params.connectivityManager))
	{}
	/** Virtual destructor. */
	virtual ~AbstractChainManager() {}

	/** Called after the chain manager is created to indicate that it can make external connections, create threads and create chains.
	 * If start throws this is a fatal startup error.
	 *
	 * Default implementation is a no-op. You must override this if you need to signal startup.
	 */
	virtual void start() {}

	/** Release any resources created by the connectivity plug-in, and terminate and join any background threads.
	 *
	 * The host will call this before the process shuts down.
	 *
	 * Note that ChannelLifecycleListener notification methods will not be invoked once shutdown begins, and 
	 * shutdown() will not be called until any in-progress notifications have returned. 
	 *
	 * The default implementation is a no-op. 
	 *
	 * Implementors should usually explicitly destroy all chains created by the manager and then 
	 * terminate and join any connections and background threads before returning from shutdown(). 
	 * Any remaining chains will be automatically destroyed after the chain manager's shutdown() has 
	 * returned, but it is usually safest to destroy them explicitly as part of shutdown(). 
	 */
	virtual void shutdown() {}

	/** Allows reporting status information from this plug-in, such as connected/disconnected
	 *  status and number of messages sent/received in each direction. 
	 */
	StatusReporter &getStatusReporter() {
		return *statusReporter;
	}

	/**
	* A logger to be used by this manager for anything which needs
	* to be written to the host's log.
	*
	* The host will take care of adding the transport and manager name to the messages
	* logged using this object.
	*/
	const Logger logger;

	/**
	* The name used in the configuration for this dynamic chain manager.
	*/
	const std::string managerName; 

	/**
	* The name used in the configuration for the transport plug-in that created this chain manager.
	*/
	const std::string transportPluginName;

protected:
	// typedefs for templated classes that are useful when writing subclasses

	/** The chain that this host can create */
	typedef typename Chain<TRANSPORT>::ptr_t chain_t;
	/** The base class */
	typedef AbstractChainManager<TRANSPORT> abstract_chain_manager_t;

	/** string-> map of chain definitions */
	ChainDefinitions chainDefinitions;

	/**
	* The interface through which the host provides services required by chain managers.
	*/
	const ChainManagerHost<TRANSPORT> host;

	/** The configuration of this plug-in. The plug-in owns and is free to mutate this data structure. */
	map_t config;

	/**
	* Helper method to get the chain definition containing the transport associated with this manager, assuming there is only one.
	*/
	const ChainDefinition &getChainDefinition() {
		if(chainDefinitions.size() != 1) throw std::runtime_error("This manager only supports a single chain definition, but there was not exactly one specified.");
		return chainDefinitions.begin()->second;
	}
	/**
	* Helper method to get the chain definition containing the transport associated with this manager, assuming there is only one.
	* @param str a chainDefinitionId.
	*/
	const ChainDefinition &getChainDefinition(const std::string &str) {
		auto it = chainDefinitions.find(str);
		if(it == chainDefinitions.end()) throw std::runtime_error("Cannot find chain definition "+str);
		return it->second;
	}

private:
	std::unique_ptr<StatusReporter> statusReporter;
};


// ***********************************************************************************************



}}}} // com.softwareag.connectivity.chainmanagers

#include <sag_internal/chain_managers.hpp>


/** Registers a connectivity transport chain manager class and its associated transport class, by exporting the necessary C free functions from this library.
 *
 *	For each chain manager derived from AbstractChainManager you should call this macro once,
 *	and ensure that your class has a constructor with the same signature as 
 *	com::softwareag::connectivity::chainmanagers::AbstractChainManager(ChainManagerConstructorParameters &).
 *
 *	Do not use SAG_DECLARE_CONNECTIVITY_TRANSPORT_CLASS for a transport class that has a chain manager; the transport will be 
 *	registered automatically by this macro. 
 *
 *	If your class is in a namespace then you must call this macro in the same namespace, passing in just the base name of the class.
 *	@since 10.2.0.0
 */
#define SAG_DECLARE_CONNECTIVITY_TRANSPORT_CHAIN_MANAGER_CLASS(Class) _SAG_DECLARE_CONNECTIVITY_TRANSPORT_CHAIN_MANAGER_CLASS(Class)

#endif // _SAG_CONNECTIVITY_CHAIN_MANAGERS_HPP_
