/*
 * Title:        sag_internal/chain_managers.hpp
 * Description:  Implementation of AbstractChainManager, ChannelLifeCycle, et al
 * $Copyright (c) 2017 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors.$
 * Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG
 * @Version:     $Id: chain_managers.hpp 321386 2017-12-14 10:26:04Z cat $
 */


#ifndef _SAG_INTERNAL_CHAIN_MANAGERS_HPP_
#define _SAG_INTERNAL_CHAIN_MANAGERS_HPP_

#include <functional>
#include <tuple>
#include <sag_internal/cpp_stdlib.hpp>

namespace com {
namespace softwareag {
namespace connectivity { 
namespace chainmanagers {


// ***********************************************************************************************
// converters to convert to / from ChainDefinitions
namespace converters_impl {
inline ChainDefinitions convertChainDefinitions(const map_t &chainDefinitions) {
	ChainDefinitions definitions;
	for(auto chainIt = chainDefinitions.begin(); chainIt != chainDefinitions.end(); chainIt++) {
		const list_t &pluginList = get<list_t>(chainIt.value());
		ChainDefinition pluginConfigs;
		for(auto pluginIt = pluginList.begin(); pluginIt != pluginList.end(); pluginIt++) {
			const map_t &pluginMap = get<map_t>(*pluginIt);
			if(pluginMap.size() != 1) throw std::runtime_error("expected exactly one map entry for a plugin (the key, value would be the name and config of the plugin)");
			auto pluginEntry = pluginMap.begin();
			pluginConfigs.push_back(PluginConfiguration(get<const char *>(pluginEntry.key()), get<map_t>(pluginEntry.value())));
		}
		if(pluginConfigs.size() < 2) throw std::runtime_error("expected at least two plugins in a chain definition");
		std::string chainName = get<const char*>(chainIt.key());
		definitions[chainName] = std::move(pluginConfigs);
	}
	return definitions;
}

inline list_t convertChainDefinition(const ChainDefinition &chainDefinition) {
	list_t list;
	for(auto it=chainDefinition.begin(); it != chainDefinition.end(); it++) {
		map_t entry;
		entry.insert(data_t(it->getPluginName()), data_t(it->getConfiguration().copy()));
		list.push_back(std::move(std::move(entry)));
	}
	return list;
}

}

namespace impl {

	template<typename TRANSPORT, typename TUPLE, size_t...SEQ> TRANSPORT* createTransport(const ManagedTransportConstructorParameters& params, 
											      const TUPLE& tuple, ampl::index_tuple<SEQ...>) {
		return new TRANSPORT(params, std::get<SEQ>(tuple)...);
	}
}

// ***********************************************************************************************
// ChainManagerHost

template<typename TRANSPORT>
template<typename... ARGS>
inline typename Chain<TRANSPORT>::ptr_t ChainManagerHost<TRANSPORT>::createChain(const std::string &chainInstanceIdSuffix, const std::string &defaultChannelTowardsHost, const list_t &channels, const ChainDefinition &chainDefinition, const map_t &substitutions, ARGS... args) const
{
	auto tuple = std::make_tuple(args...);
	return createChainImpl( chainInstanceIdSuffix, defaultChannelTowardsHost,
				channels, chainDefinition, substitutions,
				tuple);
}

template<typename TRANSPORT>
template<typename TUPLE>
typename Chain<TRANSPORT>::ptr_t ChainManagerHost<TRANSPORT>::createChainImpl(const std::string &chainInstanceIdSuffix, const std::string &defaultChannelTowardsHost, 
									      const list_t &channels, const ChainDefinition &chainDefinition, const map_t &substitutions, 
									      const TUPLE& tuple) const {
	std::string chainId = chainManager->managerName + "-" + chainInstanceIdSuffix;
	std::string managerName = chainManager->managerName; // manager can be deleted by the end of sag_create_chain
	list_t chainDefinitionList = converters_impl::convertChainDefinition(chainDefinition);
	std::function<TRANSPORT *(const ManagedTransportConstructorParameters&)> fun = [&tuple] (const ManagedTransportConstructorParameters&params) -> TRANSPORT* {
		return impl::createTransport<TRANSPORT>(params, tuple, typename ampl::make_index_tuple<std::tuple_size<TUPLE>::value>::type());
	};
	sag_plugin_t plugin = sag_create_chain(connectivityManager, chainId.c_str(), defaultChannelTowardsHost.c_str(), chainManager->managerName.c_str(), reinterpret_cast<const sag_underlying_vector_t&>(channels), reinterpret_cast<const sag_underlying_vector_t&>(chainDefinitionList), reinterpret_cast<const sag_underlying_map_t&>(substitutions), (void*) &fun);
	if (0 == plugin.plugin) throw std::runtime_error("An error occurred while creating chain " + chainId + " for manager \"" + managerName + "\"");
	return typename Chain<TRANSPORT>::ptr_t(new Chain<TRANSPORT>(reinterpret_cast<TRANSPORT*>(plugin.plugin), connectivityManager, chainId, chainManager->managerName));
}


template<typename TRANSPORT>
inline sag_error_t ChainManagerHost<TRANSPORT>::deleteChannelLifecycleListener(sag_plugin_t listener)
{
	// this is now a no-op - do not delete the listener as its lifetime is managed by the 
	// chainmanager (typically by having the chainmanager implement the listener interface directly)
	return SAG_ERROR_OK;
}

template<typename TRANSPORT>
inline sag_error_t ChainManagerHost<TRANSPORT>::callOnChannelCreated(const char* managerLogName, sag_plugin_t listener, const char *channel, bool transportwards)
{
	try {
		reinterpret_cast<ChannelLifecycleListener *>(listener.plugin)->onChannelCreated(channel, transportwards ? Direction::TOWARDS_TRANSPORT : Direction::TOWARDS_HOST);
		return SAG_ERROR_OK;
	} catch (const std::exception &e) {
		Logger(managerLogName).error("Exception while calling onChannelCreated method on ChannelLifecycleListener: %s", e.what());
		return SAG_ERROR_EXCEPTION;
	} catch (...) {
		Logger(managerLogName).error("Exception while calling onChannelCreated method on ChannelLifecycleListener: Unknown exception");
		return SAG_ERROR_EXCEPTION;
	} 
}

template<typename TRANSPORT>
inline sag_error_t ChainManagerHost<TRANSPORT>::callOnChannelDestroyed(const char* managerLogName, sag_plugin_t listener, const char *channel, bool transportwards)
{
	try {
		reinterpret_cast<ChannelLifecycleListener *>(listener.plugin)->onChannelDestroyed(channel, transportwards ? Direction::TOWARDS_TRANSPORT : Direction::TOWARDS_HOST);
		return SAG_ERROR_OK;
	} catch (const std::exception &e) {
		Logger(managerLogName).error("Exception while calling onChannelDestroyed method on ChannelLifecycleListener: %s", e.what());
		return SAG_ERROR_EXCEPTION;
	} catch (...) {
		Logger(managerLogName).error("Exception while calling onChannelDestroyed method on ChannelLifecycleListener: Unknown exception");
		return SAG_ERROR_EXCEPTION;
	} 
}


template<typename TRANSPORT>
inline void ChainManagerHost<TRANSPORT>::addChannelLifecycleListener(ChannelLifecycleListener *listener, const std::string &prefix) const
{
	sag_plugin_t plugin;
	plugin.plugin = listener;
	if (!listener) throw std::runtime_error("Null ChannelLifecycleListener passed to addChannelLifecycleListener for chain manager \""+chainManager->managerName+"\"");
	if (SAG_ERROR_OK != sag_add_channel_listener(chainManager->logger.getPrefix().c_str(), connectivityManager, plugin, prefix.c_str(), &deleteChannelLifecycleListener, &callOnChannelCreated, &callOnChannelDestroyed)) {
		throw std::runtime_error("An error occurred while adding channel listener for manager \""+chainManager->managerName+"\"");
	}
}

// ***********************************************************************************************
// Chain

template<typename TRANSPORT>
inline void Chain<TRANSPORT>::startImpl() const
{
	sag_plugin_t plugin;
	plugin.plugin = transport;
	if (!plugin.plugin || SAG_ERROR_OK != sag_start_chain(connectivityManager, plugin)) {
		throw std::runtime_error("An error occurred while starting chain \""+chainId+"\" for manager \""+managerName+"\"");
	}
}

template<typename TRANSPORT>
inline void Chain<TRANSPORT>::destroy()
{
	if (!transport) throw std::runtime_error("destroy() already called on Chain");
	sag_plugin_t plugin;
	plugin.plugin = transport;
	transport = 0;
	if (SAG_ERROR_OK != sag_destroy_chain(connectivityManager, plugin)) {
		throw std::runtime_error("An error occurred while destroying chain \""+chainId+"\" for manager \""+ managerName +"\"");
	}
}



}}}} // close namespaces
#endif
