#!/usr/bin/env python
# Copyright (c) 2015-2018 Software AG, Darmstadt, Germany and/or Software AG USA Inc., Reston, VA, USA, and/or its subsidiaries and/or its affiliates and/or their licensors. 
# Use, reproduction, transfer, publication or disclosure is prohibited except as specifically provided for in your License Agreement with Software AG 

import tempfile, os, binascii, time, shutil, io
import threading
import subprocess

from pysys.utils import filecopy
from pysys.utils import filereplace
from pysys.constants import *
from pysys.utils.pycompat import *
import locale

class DockerHelper(object): # TODO: find a way to exclude this from epydoc/sphinx if possible
	"""Helper class which provides common functionality needed by various docker and kubernetes tests. 
	
	This class is not intended to be used directly, but is invoked or subclassed by 
	other classes in the Apama docker framework. 
	
	@undocumented: _safelyOpenProcessOutputFile
	"""
	def __init__(self, parent, displayName):
		self.__parent = parent
		self.__displayName = displayName
	def __str__(self): return self.__displayName
	
	def _safelyOpenProcessOutputFile(self, path, **kwargs):
		""" This is for internal use only and not part of public supported API - do not use. 
		
		Opens a file which contains process stdout/err for reading. 
		
		Uses binary/byte str mode in Python 2, and on Python 3 in unicode character 
		string mode using the OS default encoding (or UTF-8 if in 
		7 bit ASCII/LANG=C locale) with errors='replace' so that we downgrade 
		gracefully if unexpected characters are found. 
		
		@param path: Path of file, relative to output dir or absolute path. 
		
		@return: The open file handle, in byte mode on Python 2 or character string 
		mode in Python 3. 
		"""
		path = os.path.join(self.__parent.output, path)
		if PY2: return open(path, **kwargs) # binary mode
		encoding = locale.getpreferredencoding()
		if encoding in ['ANSI_X3.4-1968', 'ascii', 'us-ascii']:
			# docker executables have sometimes been seen to return UTF-8 chars 
			# (e.g. for ...") even in LANG=C/ascii encoding, so since utf-8 
			# is a superset of ascii, permit any utf-8 char when reading output in 
			# ascii locales
			encoding='utf-8'
		return io.open(path, encoding=encoding, errors='replace')

class DockerImage(DockerHelper):
	"""Class representing a Docker images. Sort-of manages the lifetime of the images it wraps; in most cases, it will keep them
	around for a long time so that the Docker cache can speed up other tests on the same machine, but it knows how to GC them at a later
	date due to deterministic naming that encodes the ownership of the image.
	"""

	def __init__(self, parent, name, own=True):
		""" Create from a named image and parent testcase """
		DockerHelper.__init__(self, parent=parent, displayName=name)
		self.name = name
		self.parent = parent
		self.tag = None
		if not DockerImage.__doneGC:
			self.__gcImages()
		self.own = own
		self.parent.addResource(self)
	
	@staticmethod
	def generateUniqueName():
		""" Return a unique name, used as docker image name. The Name starts with string pysys followed by process id, localtime and random number.

		@return: Docker Image name
		"""
		return "pysys{:06d}{:011d}{}".format(os.getpid(), int(time.mktime(time.localtime())),
											 binascii.b2a_hex(os.urandom(2)).decode('utf-8'))

	@classmethod
	def fromBlank(cls, parent):
		""" For when you are generating a docker image by some external means, but want the test framework to manage its lifetime. Contains
		an auto-generated name (use getName()) for you to tag your image with.

		@param parent: Reference to the parent PySys testcase
		@return: a DockerImage
		"""
		name = DockerImage.__makeName()
		return DockerImage(parent, name, True)

	@classmethod
	def fromExisting(cls, parent, name, own=False):
		""" Wrap an existing image

		@param parent: Reference to the parent PySys testcase
		@param name: Name of the existing image
		@param own: Whether or not this object owns the image i.e. deletes it later
		"""
		return DockerImage(parent, name, own)

	""" Mutex to cover invocations of docker build. Empirical testing and https://github.com/docker/docker/issues/9656 suggest that this is
	    the right thing for performance. Further, two tests kicking off 'docker build' at the same time for the same content will fail to
		take advantage of the cache.
	"""
	__buildMutex = threading.Lock()

	@classmethod 
	def fromDockerfile(cls, parent, file, context=None, otherfiles=None, imageSubst=None, abortOnError=True, noCache=False, buildArgs=None ):
		""" Construct an image from a Dockerfile

		@param parent: Reference to the parent PySys testcase
		@param file: Absolute location of the Dockerfile
		@param context: Directory containing the build context. If it's a totally context-free Dockerfile, it'll just run in a blank directory.
		@param otherfiles: Other files to copy into the root of the context before build (e.g. a wrapper that we're packaging).
		                   If the path is relative, it's taken relative to the directory the Dockerfile is taken from. If the path is
		                   absolute, it's treated as absolute.
		@param imageSubst: Maps a nice friendly image name in the FROM of a Dockerfile to a DockerImage that should actually be used (so
		                   auto-testing doesn't have namespacing problems). It's a tuple (name, image).
		@param abortOnError: Abort the test if the build fails
		@param noCache: Whether or not to apply --no-cache to docker build
		"""
		parent.log.info('Creating image from dockerfile: %s', file)

		# Generate a unique name/tag for the image
		name = DockerImage.__makeName()
		# Copy the context into our output directory; using cp -a, because shutil.copytree changes stats on symlinks only (!?) which breaks
		# Docker's build caching
		context_tmp = os.path.join(parent.output, name + "_buildcontext")
		if context:
			parent.log.info("Copying context from %s to %s ", context, context_tmp)
			parent.startProcess(command='/bin/cp', arguments=["-a", context, context_tmp], stdout="__cp.out", stderr="__cp.err", timeout=1200)
		else:
			os.mkdir(context_tmp)
		
		
		# Copy & Preprocess the Dockerfile into the copied context
		dst_dockerfile = os.path.join(context_tmp, 'Dockerfile')
		
		substs = {}
		if imageSubst:
			substs["FROM " + imageSubst[0]] ="FROM " + imageSubst[1].getName()
			substs["ARG " + imageSubst[0].upper() + "_IMAGE=" + imageSubst[0]] = "ARG " + imageSubst[0].upper() + "_IMAGE=" + imageSubst[1].getName()
		filereplace.replace(file,dst_dockerfile, substs)
		if otherfiles:
			for otherfile in otherfiles:
				if os.path.isabs(otherfile):
					shutil.copy2(otherfile, context_tmp)
				else:
					shutil.copy2(os.path.dirname(file) + '/' + otherfile, context_tmp)

		# copy the docker file to the test output dir so we can debug any problems (though not if someone has already created one)
		if not os.path.exists(parent.output+'/Dockerfile'): shutil.copy2(dst_dockerfile, parent.output+'/Dockerfile')

		reqArgs = ["build", "--force-rm", "--no-cache=" + str(noCache).lower()]

		# additional --build-args lines will originate from the param buildArgs
		if buildArgs:
			for kv in buildArgs:
				reqArgs.append( "--build-arg" )
				reqArgs.append( kv )

		reqArgs.append( "-t" )
		reqArgs.append( name )
		reqArgs.append( context_tmp )

		# Build the image
		instance = parent.getInstanceCount("docker build")
		stdout, stderr = "docker_build_%i.out" % instance, "docker_build_%i.err" % instance
		try:
			DockerImage.__buildMutex.acquire()
			build = parent.startProcess(command=parent.project.DOCKER_EXE,
					arguments=reqArgs,
					environs={'DOCKER_HOST':parent.project.DOCKER_HOST},
					displayName="docker build", ignoreExitStatus=True,
					stdout=stdout, stderr=stderr)
		finally:
			DockerImage.__buildMutex.release()

		# Remove our copy of the context
		shutil.rmtree(context_tmp)
		if(build.exitStatus != 0 and abortOnError):
			parent.logFileContents(stderr, tail=True) or parent.logFileContents(stdout, tail=True, maxLines=30)
			parent.log.info('docker build failed - arguments: %s', reqArgs)
			parent.abort(BLOCKED, "docker build failed")

		return DockerImage(parent, name, True)
	
	def getName(self):
		""" The name/tag of this image """
		return self.tag if self.tag else self.name

	def addTag(self, tag):
		instance = self.parent.getInstanceCount("docker tag")
		with DockerImage.__buildMutex:
			self.parent.startProcess(command=self.parent.project.DOCKER_EXE,
					arguments=["tag", self.getName(), tag],
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker tag %s"%tag,
					stdout="docker_tag_%i.out"%instance, stderr="docker_tag_%i.err"%instance)
		self.tag = tag

	def push(self):
		instance = self.parent.getInstanceCount("docker push")
		with DockerImage.__buildMutex:
			try:
				self.parent.startProcess(command=self.parent.project.DOCKER_EXE,
					arguments=["--config", self.parent.project.DOCKER_CONFIG, "push", self.getName()], 
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker push", 
					stdout="docker_push_%i.out"%instance, stderr="docker_push_%i.err"%instance, ignoreExitStatus=False, abortOnError=True)
			except Exception:
				self.parent.logFileContents("docker_push_%i.err"%instance, tail=True) or self.parent.logFileContents("docker_push_%i.out"%instance, tail=True, maxLines=30)
				raise
		

	""" Class variable used to confine use of __gcImages to a single invocation of pysys, rather than once per test """
	__doneGC = False

	def __gcImages(self):
		""" Garbage-collects images that we don't need any more. Approximately, this gets rid of pysys-created
		images from pysys processes that are not currently running, as long as they are fairly old (aids in speeding up everybody's test
		runs, so that they can share images that are likely to be identical within the space of a day's development)
		"""
		DockerImage.__doneGC = True

		self.parent.startProcess(command=PROJECT.DOCKER_EXE,
				arguments=["images", "--no-trunc"], 
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				stdout="__gcImages.out", displayName="docker images")
		with self._safelyOpenProcessOutputFile("__gcImages.out") as f:
			image_lines = f.readlines()[1:] # byte string on python 2, char string on python 3

		pendingGC = []
		for line in image_lines:
			match = re.match("^pysys(\d\d\d\d\d\d)(\d\d\d\d\d\d\d\d\d\d\d)[^ ]+", line)
			if match:
				(name, pid, age) = (match.group(0), int(match.group(1)), int(time.mktime(time.localtime())) - int(match.group(2)))
				if(pid != os.getpid() and age > 7200):
					pendingGC.append(name)

		if len(pendingGC) > 0:
			self.parent.log.info("GC-ing Docker images " + str(pendingGC))
			self.parent.startProcess(command=self.parent.project.DOCKER_EXE,
					arguments=["rmi"] + pendingGC, 
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					#stdout='_docker_rm.err', stderr='_docker_rm.out',
					displayName="docker rmi")
			self.parent.log.info('Done GC-ing images, commencing test now.')
			self.parent.log.info('')

	
	@classmethod
	def __makeName(cls):
		""" Generate a unique name for an image """
		return DockerImage.generateUniqueName()

	def __del__(self):
		if self.own:
			self.parent.startProcess(command=self.parent.project.DOCKER_EXE, 
					arguments=["rmi", self.name],
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker rmi")
		if self.tag:
			self.parent.startProcess(command=self.parent.project.DOCKER_EXE,
					arguments=["rmi", self.tag],
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker rmi")

class DockerService(DockerHelper):
	"""Helper class for working with Docker services. Owns the service that it wraps, and will garbage collect it when this object is
	deleted (which will only happen when the parent testcase finishes).

	Non-returning methods are Fluent - that is, they return 'self' to allow method-chaining
	"""

	@classmethod
	def fromImage(cls, image):
		""" Prepare to create and run a service from an image

		@param image: a DockerImage to create from
		"""
		ret = DockerService(displayName='DockerService %s'%image,parent=image.parent)
		assert image.__class__.__name__=='DockerImage', type(image)
		ret.image = image
		ret.parent = image.parent
		return ret
	
	@classmethod
	def fromExisting(cls, parent, name, own=True):
		""" Wrap an existing running service

		@param parent: Reference to the parent PySys testcase
		@param name: Name of the existing image
		@param own: Whether or not this object owns the service i.e. deletes it later
		"""

		ret = DockerService(displayName='DockerService %s'%name,parent=parent)
		ret.image = None
		ret.parent = parent
		ret.name = name
		ret.own = own
		ret.started = True
		ret.__acquirePorts()
		return ret

	def __init__(self, displayName=None, parent=None):
		DockerHelper.__init__(self, parent=parent, displayName=displayName)
		self.parent=parent
		self.started = False
		self.extraArgs = []
		self.own = True
		self.totalStarts = 0 # Used to keep track of total number of times a Docker instance has been started
		                     # which helps when grepping logfiles etc.
	
	def run(self, cmd=None, extraArgs=None, foreground=False, abortOnError=True, environs=None):
		""" Run the service. The default UID for processes inside the service will be the current uid to aid with host-volume sharing

		@param cmd: Optional command + arguments list to run in the service
		@param extraArgs: list of extra arguments to pass to "docker run"
		@param foreground: True if this method should only return after the command has finished executing
		@param abortOnError: Abort the test if starting the service fails
		@param environs: A dictionary of the environment to run the process in (defaults to clean environment)
		"""
		assert(not self.started)
		self.started = True
		# Generate a unique name for the image
		self.name = DockerImage.generateUniqueName()
		# Create the service from the image
		args = ["service", "create", "--name", self.name]
		if not foreground:
			args.append('-d')
		args.extend(self.extraArgs)
		if extraArgs:
			args.extend(extraArgs)
		args.append(self.image.getName())
		if cmd:
			args.extend(cmd)
		instance = self.parent.getInstanceCount("docker service")
		env = environs or {}
		env['DOCKER_HOST'] = self.parent.project.DOCKER_HOST
		run = self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=args,
				environs=env,
				displayName="docker service",
				stdout="docker_service_%i.out" % instance, stderr="docker_service_%i.err" % instance)
		if(run.exitStatus != 0 and abortOnError):
			self.parent.abort(BLOCKED, "docker service failed")
		self.parent.addResource(self)
		if not DockerService.__doneGC:
			self.__gcServices()

		self.totalStarts = 1
		return self

	
	def log(self, logfile):
		""" Stream the docker logs (i.e. the stdout and stderr) for a service that has been run. Sends it live to two files, one for
		stdout, another for stderr.

		@param logfile: A named file prefix in the output directory. Suffixed with '.out' and '.err'
		"""
		assert(self.started)
		args = ["service", "logs", "-f", self.name]
		# Send stdout and stderr to different files as rit and jamc have both separately
		# observed logging going missing when we log to the same file.
		stdout = os.path.join(self.parent.output, logfile+".out")
		stderr = os.path.join(self.parent.output, logfile+".err")
		tail = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
				arguments=args,
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker service logs",
				state=BACKGROUND, 
				stdout=stdout, stderr=stderr)

		return self

	def __del__(self):
		if self.started and self.own:
			rm = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
					arguments=["service", "rm", self.name],
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker service rm")
			if(rm.exitStatus != 0):
				self.parent.addOutcome(BLOCKED, 'docker service rm %s failed'%self.name)

	""" Class variable used to confine use of __gcServices to a single invocation of pysys, rather than once per test """
	__doneGC = False

	def __gcServices(self):
		""" Garbage-collects old running services; the test framework is supposed to remove services on shutdown, but sometimes a
		serious fault bypasses this, so we take a look at them occasionally
		"""
		DockerService.__doneGC = True

		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["service", "ls"],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				stdout="__gcServices.out",
				displayName="docker service ls")
		with self._safelyOpenProcessOutputFile("__gcServices.out") as f:
			service_lines = f.readlines()[1:]

		pendingGC = []
		for line in service_lines:
			line = line.rstrip()
			match = re.match(".* pysys(\d\d\d\d\d\d)(\d\d\d\d\d\d\d\d\d\d\d)([a-z0-9_]*)( )*$", line)
			if match:
				name = "pysys%s%s%s" % (match.group(1), match.group(2), match.group(3))
				pid = int(match.group(1))
				age = int(time.mktime(time.localtime())) - int(match.group(2))
				if(pid != os.getpid() and age > 3600):
					pendingGC.append(name)

		if len(pendingGC) > 0:
			self.parent.log.info("GC-ing Docker services " + str(pendingGC))
			self.parent.startProcess(command=self.parent.project.DOCKER_EXE, 
					arguments=["service", "rm",] + pendingGC,
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker service rm")

	def __waitLogs(self, logfile_base, startCount = 0):
		# Check the logfile to be sure the server is ready
		self.log(logfile_base)

		if startCount != 0:
			self.totalStarts = startCount
		else:
			self.totalStarts = self.totalStarts + 1
		condition = ">%d" % (self.totalStarts-1)
		matches=self.parent.waitForSignal(file=logfile_base+'.out', expr='Realm Server Startup sequence completed', condition=condition, timeout=TIMEOUTS['WaitForSignal']*4)
		if len(matches) != self.totalStarts:
			self.parent.abort(BLOCKED, "Realm server has not started successfully. Expected %d starts, got %d (see %s.out/err)" % (self.totalStarts, len(matches), logfile_base))

class DockerSecret(DockerHelper):
	"""Helper class for working with Docker secrets. Owns the secret that it wraps, and will garbage collect it when this object is
	deleted (which will only happen when the parent testcase finishes).

	Non-returning methods are Fluent - that is, they return 'self' to allow method-chaining
	"""

	@classmethod
	def fromFile(cls, parent, name, file):
		""" Create a secret from a file
		@param parent: Reference to the parent PySys testcase
		@param file: The path to the file containing the secret contents, in whatever file format is required by the process that will use it
		@param name: the name of the secret used by Docker. This represents both the canonical name of the secret within a docker server, and the filename through which the secret can be accessed. Ensure the specified name does not conflict with other testcases, for example by using DockerImage.generateUniqueName(), e.g. name="my-secret-%s.properties"%DockerImage.generateUniqueName().
		"""
		ret = DockerSecret(parent=parent, file=os.path.join(parent.output, file), name=name)
		return ret

	@classmethod
	def fromContent(cls, parent, name, secretContents):
		""" Create a named secret containing th specified string contents
		@param parent: Reference to the parent PySys testcase
		@param secretContents: The contents of the secret as a character string, in whatever file format is expected by the process that will use it
		@param name: the name of the secret used by Docker. This represents both the canonical name of the secret within a docker server, and the filename through which the secret can be accessed. Ensure the specified name does not conflict with other testcases, for example by using DockerImage.generateUniqueName(), e.g. name="my-secret-%s.properties"%DockerImage.generateUniqueName().
		"""
		ret = DockerSecret(parent=parent, file='-', secretContents=secretContents, name=name)
		return ret

	def __init__(self,parent,file=None,secretContents=None,name=None):
		assert name, 'a name must be specified when creating a docker secret'
		self.__started = False
		DockerHelper.__init__(self, parent=parent, displayName="DockerSecret %s" % name)
		self.extraArgs = []
		self.own = True
		self.parent = parent
		self.file = file
		self.secretContents = secretContents
		self.name = name
		self.__create()
	
	def getName(self):
		""" The name of this secret """
		return self.name

	def __create(self):
		""" Create the secret
		"""
		assert(not self.__started)
		self.__started = True

		# Create the secret from the image
		args = ["secret", "create", self.name]
		env = {}
		env['DOCKER_HOST'] = self.parent.project.DOCKER_HOST

		out, err = self.parent.allocateUniqueStdOutErr('docker_secret')
		try:
			if not self.file or self.file == '-':
				command = 'echo "{}" | {} '.format(self.secretContents, PROJECT.DOCKER_EXE)
				command += ' '.join(args+['-'])
				self.parent.log.debug('Running Popen: %s', command)
				
				# Have to use Popen instead of parent.startProcess because startProcess didn't seem to work with the combination of commands
				# nb: there is no timeout here so this will block forever
				with open(out, "w+b") as fout, open(err, 'w+b') as ferr:
					# Popen commands must be unicode character strings on python3; on python 2 bytes or char strings both work
					process = subprocess.Popen(command, shell=True, stdout=fout, stderr=ferr)
					process.communicate()
					if(process.returncode != 0):
						self.parent.abort(BLOCKED, "docker secret create failed")
			else:
				run = self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
						arguments=args+[self.file],
						environs=env,
						displayName="docker secret create",
						stdout=out, stderr=err, ignoreExitStatus=True)
				if(run.exitStatus != 0):
					self.parent.abort(BLOCKED, "docker secret create failed")
		except Exception:
			self.parent.logFileContents(err)
			raise

		self.parent.addResource(self)
		return self

	def __del__(self):
		if self.__started and self.own:
			rm = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
					arguments=["secret", "rm", self.name],
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker secret rm")
			if(rm.exitStatus != 0):
				self.parent.addOutcome(BLOCKED, 'docker secret rm %s failed'%self.name)

class DockerContainer(DockerHelper):
	"""Helper class for working with Docker containers. Owns the container that it wraps, and will garbage collect it when this object is
	deleted (which will only happen when the parent testcase finishes).

	Non-returning methods are Fluent - that is, they return 'self' to allow method-chaining
	"""

	@classmethod
	def fromImage(cls, image):
		""" Prepare to create and run a container from an image

		@param image: a DockerImage to create from
		"""
		ret = DockerContainer(parent = image.parent, displayName='DockerContainer %s'%image.name)
		assert image.__class__.__name__=='DockerImage', type(image)
		ret.image = image
		return ret
	
	@classmethod
	def fromExisting(cls, parent, name, own=True):
		""" Wrap an existing running container

		@param parent: Reference to the parent PySys testcase
		@param name: Name of the existing image
		@param own: Whether or not this object owns the container i.e. deletes it later
		"""

		ret = DockerContainer(parent=parent, displayName='DockerContainer %s'%name)
		ret.image = None
		ret.name = name
		ret.own = own
		ret.started = True
		ret.__acquirePorts()
		return ret

	def __init__(self, parent=None, displayName=None):
		DockerHelper.__init__(self, parent=parent, displayName=displayName)
		self.parent = parent
		
		self.started = False
		self.extraArgs = []
		self.own = True
		self.totalStarts = 0 # Used to keep track of total number of times a Docker instance has been started
		                     # which helps when grepping logfiles etc.
	
	def run(self, cmd=None, extraArgs=None, foreground=False, abortOnError=True, environs=None):
		""" Run the container. The default UID for processes inside the container will be the current uid to aid with host-volume sharing

		@param cmd: Optional command + arguments list to run in the container
		@param extraArgs: list of extra arguments to pass to "docker run"
		@param foreground: True if this method should only return after the command has finished executing
		@param abortOnError: Abort the test if starting the container fails
		@param environs: A dictionary of the environment to run the process in (defaults to clean environment)
		"""
		assert(not self.started)
		self.started = True
		# Generate a unique name for the image
		self.name = DockerImage.generateUniqueName()
		# Create the container from the image
		args = ["run", "--name", self.name]
		if not foreground:
			args.append('-d')
		args.extend(self.extraArgs)
		if extraArgs:
			args.extend(extraArgs)
		args.append(self.image.getName())
		if cmd:
			args.extend(cmd)
		instance = self.parent.getInstanceCount("docker run")
		env = environs or {}
		env['DOCKER_HOST'] = self.parent.project.DOCKER_HOST
		run = self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=args,
				environs=env,
				displayName="docker run",
				stdout="docker_run_%i.out" % instance, stderr="docker_run_%i.err" % instance)
		if(run.exitStatus != 0 and abortOnError):
			self.parent.abort(BLOCKED, "docker run failed")
		self.parent.addResource(self)
		self.__acquirePorts()
		if not DockerContainer.__doneGC:
			self.__gcContainers()

		self.totalStarts = 1
		return self

	
	def log(self, logfile):
		""" Stream the docker logs (i.e. the stdout and stderr) for a container that has been run. Sends it live to two files, one for
		stdout, another for stderr.

		@param logfile: A named file prefix in the output directory. Suffixed with '.out' and '.err'
		"""
		assert(self.started)
		args = ["logs", "-f", self.name]
		# Send stdout and stderr to different files as rit and jamc have both separately
		# observed logging going missing when we log to the same file.
		stdout = os.path.join(self.parent.output, logfile+".out")
		stderr = os.path.join(self.parent.output, logfile+".err")
		tail = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
				arguments=args,
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker logs",
				state=BACKGROUND, 
				stdout=stdout, stderr=stderr)

		return self

	def volumeFromHost(self, hostpath, containerpath):
		""" Bind mount a directory from the host to the container i.e. -v
		Can't do this to a container that has already been started with run()

		@param hostpath: Absolute path on the host
		@param containerpath: Absolute path in the container
		"""
		assert(not self.started)
		self.extraArgs.extend(["-v", hostpath + ":" + containerpath])

		return self
	
	def exposePort(self, containerPort):
		""" Expose a port on the container to a port on the host, only for a container that hasn't been run yet

		@param containerPort: Port within the container
		"""
		assert(not self.started)
		hostPort = self.parent.getNextAvailableTCPPort()
		self.extraArgs.extend(["-p", "%i:%i" % (hostPort, containerPort)])

		return self
	
	def getExternalPort(self, port):
		""" Return the external port for a port inside the container

		@param port: The port inside the container
		@return: The port on the host operating system
		"""
		return self.portsContainerToHost[port]

	def commit(self):
		""" Do a 'docker commit' on this container

		@return: A DockerImage covering the newly created image
		"""
		img = DockerImage.fromBlank(self.parent)
		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["commit", self.name, img.getName()],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				stdout="__commit.out", stderr="__commit.err",
				displayName="docker commit")
		return img
	
	def start(self, startCount = 0):
		""" Do a 'docker start' on this container (starts a stopped container)"""
		instance = self.parent.getInstanceCount("docker start")
		logfile_base = "docker_start_%i" % instance
		start = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
				arguments=["start", self.name],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker start",
				stdout=logfile_base+'.cmd.out', stderr=logfile_base+'.cmd.err')

		if(start.exitStatus != 0):
			self.parent.abort(BLOCKED, "docker start failed")

		self.started = True
		self.__waitLogs(logfile_base, startCount)
		return self

	def stop(self):
		""" Do a 'docker stop' on this container """
		stop = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
				arguments=["stop", "--time=60", self.name],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker stop",
				stdout="__stop.out", stderr="__stop.err")
		if(stop.exitStatus != 0):
			self.parent.abort(BLOCKED, "docker stop failed")
		self.started = False
		
		return self

	def restart(self, startCount = 0):
		""" Do a 'docker restart' on this container """
		instance = self.parent.getInstanceCount("docker restart")
		logfile_base = "docker_restart_%i" % instance
		restart = self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["restart", "--time=60", self.name],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker restart",
				stdout=logfile_base+'.cmd.out', stderr=logfile_base+'.cmd.err')
		if(restart.exitStatus != 0):
			self.parent.abort(BLOCKED, "docker restart failed")

		self.started = True
		self.__waitLogs(logfile_base, startCount)

		return self
	
	def wait(self):
		""" Block until this container has exited
		
		@return: The container's return code
		"""
		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["wait", self.name], 
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker wait",
				stdout="__wait.out", stderr="__wait.err",
				timeout=TIMEOUTS['WaitForSocket'])

		with self._safelyOpenProcessOutputFile('__wait.out') as f:
			wait_out = f.read()
			if len(wait_out) > 0:
				return int(wait_out)

		return None
	
	def cp(self, file):
		""" 'docker cp' a file out of the container into your output directory
		
		@param file: Absolute path of the file in the container. The file in your output directory will share its basename.
		"""
		cp = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
		                              arguments=["cp", self.name + ':' + file, self.parent.output],
												environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
		                              displayName="docker cp", stdout="__dockercp.out", stderr="__dockercp.err")
		if(cp.exitStatus != 0):
			self.parent.abort(BLOCKED, "docker cp failed")

		return self

	def __del__(self):
		if self.started and self.own:
			rm = self.parent.startProcess(command=PROJECT.DOCKER_EXE,
					arguments=["rm", "-f", "-v", self.name],
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker rm %s"%self.name)
			if(rm.exitStatus != 0):
				self.parent.addOutcome(BLOCKED, 'docker rm %s failed' %self.name)

	""" Class variable used to confine use of __gcContainers to a single invocation of pysys, rather than once per test """
	__doneGC = False

	def __gcContainers(self):
		""" Garbage-collects old running containers; the test framework is supposed to remove containers on shutdown, but sometimes a
		serious fault bypasses this, so we take a look at them occasionally
		"""
		DockerContainer.__doneGC = True

		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["ps", "-a"],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				stdout="__gcContainers.out",
				displayName="docker ps")
		with self._safelyOpenProcessOutputFile("__gcContainers.out") as f:
			container_lines = f.readlines()[1:] # bytes on python 2, characters on python 3

		pendingGC = []
		for line in container_lines:
			line = line.rstrip()
			match = re.match(".* pysys(\d\d\d\d\d\d)(\d\d\d\d\d\d\d\d\d\d\d)([a-z0-9_]*)( )*$", line)
			if match:
				name = "pysys%s%s%s" % (match.group(1), match.group(2), match.group(3))
				pid = int(match.group(1))
				age = int(time.mktime(time.localtime())) - int(match.group(2))
				if(pid != os.getpid() and age > 3600):
					pendingGC.append(name)

		if len(pendingGC) > 0:
			self.parent.log.info("GC-ing Docker containers " + str(pendingGC))
			self.parent.startProcess(command=self.parent.project.DOCKER_EXE, 
					arguments=["stop", "-t", "60"] + pendingGC,
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker stop")
			self.parent.startProcess(command=self.parent.project.DOCKER_EXE, 
					arguments=["rm", "-f", "-v"] + pendingGC,
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker rm")

	def __acquirePorts(self):
		""" Calculates all ports that are exposed to the host by this container, populating the dictionary self.portsContainerToHost """
		self.portsContainerToHost = {}
		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["port", self.name],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				stdout="__ports.out", 
				displayName="docker port")
		with self._safelyOpenProcessOutputFile("__ports.out") as f:
			for line in f.readlines(): # bytes on python 2, chars on python 3
				m = re.match("^([\d]+)\/tcp -> .*:([\d]+)", line)
				self.portsContainerToHost[int(m.group(1))] = int(m.group(2))

	def __waitLogs(self, logfile_base, startCount = 0):
		# Check the logfile to be sure the server is ready
		self.log(logfile_base)

		if startCount != 0:
			self.totalStarts = startCount
		else:
			self.totalStarts = self.totalStarts + 1
		condition = ">%d" % (self.totalStarts-1)
		matches=self.parent.waitForSignal(file=logfile_base+'.out', expr='Realm Server Startup sequence completed', condition=condition, timeout=TIMEOUTS['WaitForSignal']*4)
		if len(matches) != self.totalStarts:
			self.parent.abort(BLOCKED, "Realm server has not started successfully. Expected %d starts, got %d (see %s.out/err)" % (self.totalStarts, len(matches), logfile_base))



class DockerSwarm(DockerHelper):
	""" Helper class for working with docker-compose environments. As with the other Docker classes, it attempts to clean up after itself
	"""

	def __init__(self, parent, file, imageSubst=None, externalPorts=None, hostPathSubst=None, buildSubst=None, additionalFiles=None,
	             otherSubsts=None):
		""" Create from a docker-compose file. We're going to be testing sample files that have hardcoded aspects to them; exposing ports to
		fixed external ports; naming specific images from earlier steps; refering a specific path to build a Dockerfile in. However, this is
		inappropriate for auto-tests, so this method will tailor the file according to substitutions.

		@param parent: Reference to the parent PySys testcase
		@param file: Absolute path to the docker-compose file
		@param imageSubst: dictionary from fixed image names to DockerImages that should be used instead
		@param externalPorts: a list of external port numbers mentioned in this compose that get randomly reallocated. Only the DockerContainer 
		                      will know which to.
		@param hostPathSubst: dictionary from the host path in a volume declaration, to the real host path we'd like it to be
		@param buildSubst: dictionary from a build: path to a (directory of real context, [Dockerfile names within context])
		@param additionalFiles: dictionary from a build path to [absolute paths of additional files to copy into the context]
		                        You might use this to copy Dockerfiles into the context if they're not already there
		@param otherSubsts: Any other kind of arbitrary textual substitution not covered by the above options. Mapping of string to
		                    replacement string.
		"""
		DockerHelper.__init__(self, parent=parent, displayName='DockerSwarm %s'%os.path.basename(file))
		if not imageSubst:
			imageSubst = {}
		if not externalPorts:
			externalPorts = []
		if not hostPathSubst:
			hostPathSubst = {}
		if not buildSubst:
			buildSubst = {}
		if not otherSubsts:
			otherSubsts = {}
		self.parent = parent
		self.isUp = False
		parent.log.info("Tailoring a docker-compose file from " + file)
		self.tailoredFile = os.path.join(parent.output, "docker-compose-%s.yml" % binascii.b2a_hex(os.urandom(8)))

		# Create a globally unique project name
		self.name = DockerImage.generateUniqueName()

		# Perform substitutions, in the compose and in any referenced Dockerfiles
		replaceMap = {}
		dfReplaceMap = {}
		for i in imageSubst:
			assert imageSubst[i].__class__.__name__=='DockerImage', type(imageSubst[i])
			replaceMap["image: " + i] = "image: " + imageSubst[i].getName()
			dfReplaceMap["FROM " + i] = "FROM " + imageSubst[i].getName()
			dfReplaceMap["ARG " + i.upper() + "_IMAGE=" + i] = "ARG " + i.upper() + "_IMAGE=" + imageSubst[i].getName()
		for i in externalPorts:
			assert(isinstance(i, int))
			replaceMap["- \"%s:" % i] = "- \""
			replaceMap["- %s:" % i] = "- "
		for i in hostPathSubst:
			assert(isstring(i) and isstring(hostPathSubst[i]))
			replaceMap["- %s:" % i] = "- %s:" % hostPathSubst[i]
		for i in otherSubsts:
			replaceMap[i] = otherSubsts[i]

		j = 0
		self.buildContexts = []
		for i in buildSubst:
			# Set up the copy of the build context, including the additional files
			context_tmp = os.path.join(parent.output, "%s_%i_buildcontext" % (self.name, j))
			(context, dockerfiles) = buildSubst[i]
			parent.startProcess(command='/bin/cp', arguments=["-a", context, context_tmp], stdout="__cp.out", stderr="__cp.err")

			if additionalFiles and i in additionalFiles:
				for f in additionalFiles.get(i):
					shutil.copy2(f, os.path.join(context_tmp, os.path.basename(f)))

			# Tailor any Dockerfiles, and tailor the compose file to point to this copy of the build context
			for df in dockerfiles:
				filereplace.replace(os.path.join(context_tmp, df), os.path.join(context_tmp, df + "_"), dfReplaceMap)
				os.rename(os.path.join(context_tmp, df + "_"), os.path.join(context_tmp, df))
			replaceMap["build: " + i] = "build: " + context_tmp
			replaceMap["context: " + i] = "context: " + context_tmp
			self.buildContexts.append(context_tmp)

			j = j + 1
			
		filereplace.replace(file, self.tailoredFile, replaceMap)

		self.commonArgs = ["-c", self.tailoredFile]
	
	""" Mutex to cover invocations of docker stack deploy, to indirectly limit the number of simultaneous builds. See
	    DockerImage.__buildMutex. N.B. possibly redundant in stack because build must be done explicitly
	"""
	__deployMutex = threading.Lock()

	def deploy(self, abortOnError=True, environs=None):
		""" Bring up all services in the environment

		@param abortOnError: Abort the test if this fails
		@param environs: A dictionary of the environment to run the process in (defaults to clean environment)
		"""
		assert(not self.isUp)
		self.isUp = True
		instance = self.parent.getInstanceCount("docker stack deploy")
		try:
			DockerSwarm.__deployMutex.acquire()
			env = environs or {}
			env['DOCKER_HOST'] = self.parent.project.DOCKER_HOST
			deploy = self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
					arguments= ["stack", "deploy"] + self.commonArgs + [ self.name ] ,
					environs=env,
					displayName="docker stack deploy",
					stdout=self.parent.output + "/docker_stack_deploy_%i.out" % instance,
					stderr=self.parent.output + "/docker_stack_deploy_%i.err" % instance)
		finally:
			DockerSwarm.__deployMutex.release()
			
		if(deploy.exitStatus != 0 and abortOnError):
			self.parent.abort(BLOCKED, "docker stack deploy failed")
		self.parent.addResource(self)
	
	def getContainers(self):
		""" Returns all the service containers for a running environment

		@return: A map from service name to a list of all DockerContainers running that service
		"""
		assert(self.isUp)

		stdout = os.path.join(self.parent.output, "__getcontainers.out")
		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["ps", "-a"], 
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker ps", 
				stdout=stdout)
		with self._safelyOpenProcessOutputFile(stdout) as f:
			lines = f.readlines() # bytes on python 2, chars on python 3

		ret = {}
		for line in lines:
			m = re.match(".* " + self.name + "_(.*)\.([0-9]+)\.(.*)", line)
			if m:
				serviceName = m.group(1)
				serviceNumber = m.group(2)
				serviceId = m.group(3)
				if serviceName not in ret:
					ret[serviceName] = []
				containerName = self.name + "_" + serviceName + "." + serviceNumber +"."+serviceId
				ret[serviceName].append(DockerContainer.fromExisting(self.parent, containerName, own=False))

		return ret

	def rm(self, services=None):
		""" docker-compose rm

		@param services: Names of services to rm. If none provided, removes them all.
		"""
		myServices = [] if not services else services
		rm = self.parent.startProcess(PROJECT.DOCKER_EXE,
				arguments=["stack","rm",self.name],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker stack rm",			
				stdout=self.parent.output + "/docker_stack_rm_%s.out" % self.name,
				stderr=self.parent.output + "/docker_stack_rm_%s.err" % self.name)
		if(rm.exitStatus != 0):
			self.parent.abort(BLOCKED, "docker stack rm failed")
		# If removing all services, also clean up Volumes and Networks
		if not services:
			self.__gcImp(self.name)

	def getVolumes(self, name=None):
		""" Returns all the volumes for a running environment

		@return: A map from volume name to tuple of (driver type, full name)
		"""
		if not name:
			name = self.name
		stdout = os.path.join(self.parent.output, "__getvolumes.out")
		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["volume", "ls"], 
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker volume", 
				stdout=stdout)
		with self._safelyOpenProcessOutputFile(stdout) as f:
			lines = f.readlines() # bytes on python 2, chars on python 3

		ret = {}
		for line in lines:
			m = re.match("(\w*)\s*" + name + "_(.*)", line)
			if m:
				volumeDriver = m.group(1)
				volumeName = m.group(2)
				m2 = re.match("(\w*)\s*(\w*)", line)
				ret[volumeName] = (volumeDriver, m2.group(2))

		return ret

	def getNetworks(self, name=None):
		""" Returns all the networks for a running environment

		@return: A map from network name to tuple of (driver type, full name, id)
		"""
		if not name:
			name = self.name
		stdout = os.path.join(self.parent.output, "__getnetworks.out")
		self.parent.startProcess(command=PROJECT.DOCKER_EXE,
				arguments=["network", "ls"],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker network",
				stdout=stdout)
		with self._safelyOpenProcessOutputFile(stdout) as f:
			lines = f.readlines() # bytes on python 2, chars on python 3

		ret = {}
		for line in lines:
			m = re.match("(\w*)\s*" + name + "_(\w*)\s*(\w*)", line)
			if m:
				networkId = m.group(1)
				networkName = m.group(2)
				networkDriver = m.group(3)
				m2 = re.match("(\w*)\s*(\w*)", line)
				ret[networkName] = (networkDriver, m2.group(2), networkId)

		return ret

	def __del__(self):
		if self.isUp:
			self.rm()
			# Remove our copies of build contexts
			for i in self.buildContexts:
				shutil.rmtree(i)
		self.__gcImp(self.name)
		if not DockerSwarm.__doneGC:
			self.__gc()

	""" Class variable used to confine use of __gc to a single invocation of pysys, rather than once per test """
	__doneGC = False
	
	def __gc(self):
		""" Garbage-collects old volumes and networks; the test framework is supposed to remove networks and volumes on shutdown, but sometimes a
		serious fault bypasses this, so we take a look at them occasionally
		"""
		DockerSwarm.__doneGC = True
		self.__gcImp("pysys(\d\d\d\d\d\d)(\d\d\d\d\d\d\d\d\d\d\d)([a-z0-9_]*)")

	def __gcImp(self, name):
		""" Garbage-collects volumes and networks for a specific name
		"""
		volumes = self.getVolumes(name)
		volumeGC = []
		for vol in volumes:
			volType, volName = volumes[vol]
			volumeGC.append(volName)
		networks = self.getNetworks(name)
		networkGC = []
		for net in networks:
			netType, netName, netId = networks[net]
			networkGC.append(netName)

		if len(volumeGC) > 0:
			self.parent.log.info("GC-ing Docker volumes " + str(volumeGC))
			self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
					arguments=["volume", "rm"] + volumeGC, 
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker volume rm", 
					stdout='__gcVolumes.out')

		if len(networkGC) > 0:
			self.parent.log.info("GC-ing Docker networks " + str(networkGC))
			self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
					arguments=["network", "rm"] + networkGC,
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker volume rm",
					stdout='__gcNeworks.out')



class DockerComposition(DockerHelper):
	""" Helper class for working with docker-compose environments. As with the other Docker classes, it attempts to clean up after itself
	"""

	def __init__(self, parent, file, imageSubst=None, externalPorts=None, hostPathSubst=None, buildSubst=None, additionalFiles=None,
	             otherSubsts=None):
		""" Create from a docker-compose file. We're going to be testing sample files that have hardcoded aspects to them; exposing ports to
		fixed external ports; naming specific images from earlier steps; refering a specific path to build a Dockerfile in. However, this is
		inappropriate for auto-tests, so this method will tailor the file according to substitutions.

		@param parent: Reference to the parent PySys testcase
		@param file: Absolute path to the docker-compose file
		@param imageSubst: dictionary from fixed image names to DockerImages that should be used instead
		@param externalPorts: a list of external port numbers mentioned in this compose that get randomly reallocated. Only the DockerContainer 
		                      will know which to.
		@param hostPathSubst: dictionary from the host path in a volume declaration, to the real host path we'd like it to be
		@param buildSubst: dictionary from a build: path to a (directory of real context, [Dockerfile names within context])
		@param additionalFiles: dictionary from a build path to [absolute paths of additional files to copy into the context]
		                        You might use this to copy Dockerfiles into the context if they're not already there
		@param otherSubsts: Any other kind of arbitrary textual substitution not covered by the above options. Mapping of string to
		                    replacement string.
		"""
		#TODO - substitute user: if necessary
		DockerHelper.__init__(self, parent=parent, displayName='DockerComposition %s'%os.path.basename(file))
		if not imageSubst:
			imageSubst = {}
		if not externalPorts:
			externalPorts = []
		if not hostPathSubst:
			hostPathSubst = {}
		if not buildSubst:
			buildSubst = {}
		if not otherSubsts:
			otherSubsts = {}
		self.parent = parent
		self.isUp = False
		parent.log.info("Tailoring a docker-compose file from " + file)
		self.tailoredFile = os.path.join(parent.output, "docker-compose-%s.yml" % binascii.b2a_hex(os.urandom(8)))

		# Create a globally unique project name
		self.name = DockerImage.generateUniqueName()

		# Perform substitutions, in the compose and in any referenced Dockerfiles
		replaceMap = {}
		dfReplaceMap = {}
		for i in imageSubst:
			assert imageSubst[i].__class__.__name__=='DockerImage', type(imageSubst[i])
			replaceMap["image: " + i] = "image: " + imageSubst[i].getName()
			dfReplaceMap["FROM " + i] = "FROM " + imageSubst[i].getName()
			dfReplaceMap["ARG " + i.upper() + "_IMAGE=" + i] = "ARG " + i.upper() + "_IMAGE=" + imageSubst[i].getName()
		for i in externalPorts:
			assert(isinstance(i, int))
			replaceMap["- \"%s:" % i] = "- \""
			replaceMap["- %s:" % i] = "- "
		for i in hostPathSubst:
			assert(isstring(i) and isstring(hostPathSubst[i]))
			replaceMap["- %s:" % i] = "- %s:" % hostPathSubst[i]
		for i in otherSubsts:
			replaceMap[i] = otherSubsts[i]

		j = 0
		self.buildContexts = []
		for i in buildSubst:
			# Set up the copy of the build context, including the additional files
			context_tmp = os.path.join(parent.output, "%s_%i_buildcontext" % (self.name, j))
			(context, dockerfiles) = buildSubst[i]
			parent.startProcess(command='/bin/cp', arguments=["-a", context, context_tmp], stdout="__cp.out", stderr="__cp.err")

			if additionalFiles and i in additionalFiles:
				for f in additionalFiles.get(i):
					shutil.copy2(f, os.path.join(context_tmp, os.path.basename(f)))

			# Tailor any Dockerfiles, and tailor the compose file to point to this copy of the build context
			for df in dockerfiles:
				filereplace.replace(os.path.join(context_tmp, df), os.path.join(context_tmp, df + "_"), dfReplaceMap)
				os.rename(os.path.join(context_tmp, df + "_"), os.path.join(context_tmp, df))
			replaceMap["build: " + i] = "build: " + context_tmp
			replaceMap["context: " + i] = "context: " + context_tmp
			self.buildContexts.append(context_tmp)

			j = j + 1
			
		filereplace.replace(file, self.tailoredFile, replaceMap)

		self.commonArgs = ["-p", self.name, "-f", self.tailoredFile]
	
	""" Mutex to cover invocations of docker-compose up, to indirectly limit the number of simultaneous builds. See
	    DockerImage.__buildMutex.
	"""
	__upMutex = threading.Lock()

	def up(self, abortOnError=True, environs=None):
		""" Bring up all services in the environment

		@param abortOnError: Abort the test if this fails
		@param environs: A dictionary of the environment to run the process in (defaults to clean environment)
		"""
		assert(not self.isUp)
		self.isUp = True
		instance = self.parent.getInstanceCount("docker-compose up")
		stdout, stderr = self.parent.output + "/docker_compose_up_%i.out" % instance, self.parent.output + "/docker_compose_up_%i.err" % instance
		try:
			DockerComposition.__upMutex.acquire()
			env = environs or {}
			env['DOCKER_HOST'] = self.parent.project.DOCKER_HOST
			up = self.parent.startProcess(command=PROJECT.DOCKER_COMPOSE_EXE, 
					arguments=self.commonArgs + ["up", "-d", "--no-recreate"],
					environs=env,
					displayName="docker-compose up", ignoreExitStatus=True, 
					stdout=stdout, stderr=stderr)
		finally:
			DockerComposition.__upMutex.release()
			
		if(up.exitStatus != 0 and abortOnError):
			self.parent.logFileContents(stderr, includes=['ERROR:']) or self.parent.logFileContents(stderr)
			self.parent.abort(BLOCKED, "docker-compose up failed")
		self.parent.addResource(self)
	
	def getContainers(self):
		""" Returns all the service containers for a running environment

		@return: A map from service name to a list of all DockerContainers running that service
		"""
		assert(self.isUp)

		stdout = os.path.join(self.parent.output, "__getcontainers.out")
		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["ps", "-a"], 
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker ps", 
				stdout=stdout)
		with self._safelyOpenProcessOutputFile(stdout) as f:
			lines = f.readlines() # bytes on python 2, chars on python 3

		ret = {}
		for line in lines:
			m = re.match(".* " + self.name + "_(.*)_([0-9]+)", line)
			if m:
				serviceName = m.group(1)
				serviceNumber = m.group(2)
				if serviceName not in ret:
					ret[serviceName] = []
				containerName = self.name + "_" + serviceName + "_" + serviceNumber
				ret[serviceName].append(DockerContainer.fromExisting(self.parent, containerName, own=False))

		return ret
	
	def getVolumes(self, name=None):
		""" Returns all the volumes for a running environment

		@return: A map from volume name to tuple of (driver type, full name)
		"""
		if not name:
			name = self.name
		stdout = os.path.join(self.parent.output, "__getvolumes.out")
		self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
				arguments=["volume", "ls"], 
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker volume", 
				stdout=stdout)
		with self._safelyOpenProcessOutputFile(stdout) as f:
			lines = f.readlines() # bytes on python 2, chars on python 3

		ret = {}
		for line in lines:
			m = re.match("(\w*)\s*" + name + "_(.*)", line)
			if m:
				volumeDriver = m.group(1)
				volumeName = m.group(2)
				m2 = re.match("(\w*)\s*(\w*)", line)
				ret[volumeName] = (volumeDriver, m2.group(2))

		return ret

	def getNetworks(self, name=None): # TODO: remove duplication of this method on multiple classes
		""" Returns all the networks for a running environment

		@return: A map from network name to tuple of (driver type, full name, id)
		"""
		if not name:
			name = self.name
		stdout = os.path.join(self.parent.output, "__getnetworks.out")
		self.parent.startProcess(command=PROJECT.DOCKER_EXE,
				arguments=["network", "ls"],
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker network",
				stdout=stdout)
		with self._safelyOpenProcessOutputFile(stdout) as f:
			lines = f.readlines()

		ret = {}
		for line in lines:
			m = re.match("(\w*)\s*" + name + "_(\w*)\s*(\w*)", line)
			if m:
				networkId = m.group(1)
				networkName = m.group(2)
				networkDriver = m.group(3)
				m2 = re.match("(\w*)\s*(\w*)", line)
				ret[networkName] = (networkDriver, m2.group(2), networkId)

		return ret

	def stop(self):
		""" docker-compose stop """
		stop = self.parent.startProcess(PROJECT.DOCKER_COMPOSE_EXE,
				arguments=self.commonArgs + ["stop"], 
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker-compose stop")
		if(stop.exitStatus != 0):
			self.parent.abort(BLOCKED, "docker-compose stop failed")
		self.isUp = False
	
	def rm(self, services=None):
		""" docker-compose rm

		@param services: Names of services to rm. If none provided, removes them all.
		"""
		myServices = [] if not services else services
		rm = self.parent.startProcess(PROJECT.DOCKER_COMPOSE_EXE,
				arguments=self.commonArgs + ["rm", "-v", "--force"] + myServices,
				environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
				displayName="docker-compose rm")
		if(rm.exitStatus != 0):
			self.parent.abort(BLOCKED, "docker-compose rm failed")
		# If removing all services, also clean up Volumes and Networks
		if not services:
			self.__gcImp(self.name)

	def __del__(self):
		if self.isUp:
			self.stop()
			self.rm()
			# Remove our copies of build contexts
			for i in self.buildContexts:
				shutil.rmtree(i)
		self.__gcImp(self.name)
		if not DockerComposition.__doneGC:
			self.__gc()

	""" Class variable used to confine use of __gc to a single invocation of pysys, rather than once per test """
	__doneGC = False
	
	def __gc(self):
		""" Garbage-collects old volumes and networks; the test framework is supposed to remove networks and volumes on shutdown, but sometimes a
		serious fault bypasses this, so we take a look at them occasionally
		"""
		DockerComposition.__doneGC = True
		self.__gcImp("pysys(\d\d\d\d\d\d)(\d\d\d\d\d\d\d\d\d\d\d)([a-z0-9_]*)")

	def __gcImp(self, name):
		""" Garbage-collects volumes and networks for a specific name
		"""
		volumes = self.getVolumes(name)
		volumeGC = []
		for vol in volumes:
			volType, volName = volumes[vol]
			volumeGC.append(volName)
		networks = self.getNetworks(name)
		networkGC = []
		for net in networks:
			netType, netName, netId = networks[net]
			networkGC.append(netName)

		if len(volumeGC) > 0:
			self.parent.log.info("GC-ing Docker volumes " + str(volumeGC))
			self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
					arguments=["volume", "rm"] + volumeGC, 
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker volume rm", 
					stdout='__gcVolumes.out')

		if len(networkGC) > 0:
			self.parent.log.info("GC-ing Docker networks " + str(networkGC))
			self.parent.startProcess(command=PROJECT.DOCKER_EXE, 
					arguments=["network", "rm"] + networkGC,
					environs={'DOCKER_HOST':self.parent.project.DOCKER_HOST},
					displayName="docker volume rm",
					stdout='__gcNeworks.out')

